from .ffterm import *
from .forcefield import ForceField

__all__ = ['dff_fuzzy_match']


def dff_fuzzy_match(term, ff):
    '''
    Match a approximate bonded term from a force field by the atom type string.

    It is a implementation of the `parameter transfer` in DFF.

    Parameters
    ----------
    term : BondTerm or AngleTerm or DihedralTerm or ImproperTerm or ChargeIncrementTerm
    ff : ForceField

    Returns
    -------
    matched_term : the same as term
    score : int
    '''
    max_score = 5
    best_score = 0
    best_match = None
    term_candidates = {
        BondTerm           : ff.bond_terms,
        AngleTerm          : ff.angle_terms,
        DihedralTerm       : ff.dihedral_terms,
        ImproperTerm       : ff.improper_terms,
        ChargeIncrementTerm: ff.qinc_terms,
    }
    for term_type, candidates in term_candidates.items():
        if type(term) is term_type:
            for candidate in candidates.values():
                score = _dff_fuzzy_compare(term, candidate, max_score)
                if score > best_score:
                    best_score = score
                    best_match = candidate
            return best_match, best_score

    raise Exception('Invalid type for term')


def _dff_fuzzy_compare(term, candidate, max_score):
    if type(term) is ChargeIncrementTerm and isinstance(candidate, ChargeIncrementTerm):
        score1 = _dff_fuzzy_score(term.type1, candidate.type1, max_score, 3)
        score2 = _dff_fuzzy_score(term.type2, candidate.type2, max_score, 3)
        if score1 * score2 == 0:
            return 0
        return score1 + score2
    if type(term) is BondTerm and isinstance(candidate, BondTerm):
        score1 = _dff_fuzzy_score(term.type1, candidate.type1, max_score, 3)
        score2 = _dff_fuzzy_score(term.type2, candidate.type2, max_score, 3)
        if score1 * score2 == 0:
            return 0
        return score1 + score2
    elif type(term) is AngleTerm and isinstance(candidate, AngleTerm):
        score1 = _dff_fuzzy_score(term.type1, candidate.type1, max_score, 2)
        score2 = _dff_fuzzy_score(term.type2, candidate.type2, max_score, 3)
        score3 = _dff_fuzzy_score(term.type3, candidate.type3, max_score, 2)
        if score1 * score2 * score3 == 0:
            return 0
        return score1 + score2 + score3
    elif type(term) is DihedralTerm and isinstance(candidate, DihedralTerm):
        score1 = _dff_fuzzy_score(term.type1, candidate.type1, max_score, 2)
        score2 = _dff_fuzzy_score(term.type2, candidate.type2, max_score, 3)
        score3 = _dff_fuzzy_score(term.type3, candidate.type3, max_score, 3)
        score4 = _dff_fuzzy_score(term.type4, candidate.type4, max_score, 2)
        if score1 == 0 and (term.type1 == '*' or candidate.type1 == '*'):
            score1 = 1
        if score4 == 0 and (term.type4 == '*' or candidate.type4 == '*'):
            score4 = 1
        if score1 * score2 * score3 * score4 == 0:
            return 0
        return score1 + score2 + score3 + score4
    elif type(term) is ImproperTerm and isinstance(candidate, ImproperTerm):
        score1 = _dff_fuzzy_score(term.type1, candidate.type1, max_score, 3)
        score2 = _dff_fuzzy_score(term.type2, candidate.type2, max_score, 2)
        score3 = _dff_fuzzy_score(term.type3, candidate.type3, max_score, 2)
        score4 = _dff_fuzzy_score(term.type4, candidate.type4, max_score, 2)
        if score2 == 0 and (term.type2 == '*' or candidate.type2 == '*'):
            score2 = 1
        if score3 == 0 and (term.type3 == '*' or candidate.type3 == '*'):
            score3 = 1
        if score4 == 0 and (term.type4 == '*' or candidate.type4 == '*'):
            score4 = 1
        if score1 * score2 * score3 * score4 == 0:
            return 0
        return score1 + score2 + score3 + score4


def _dff_fuzzy_score(name1, name2, max_score, min_length):
    len1, len2 = len(name1), len(name2)
    if len1 < min_length or len2 < min_length:
        return 0
    if name1 == name2:
        return max_score
    elif len1 == len2:
        return _dff_fuzzy_score(name1[:-1], name2[:-1], max_score - 2, min_length)
    elif len1 > len2:
        return _dff_fuzzy_score(name1[:len2], name2, max_score - (len1 - len2), min_length)
    else:
        return _dff_fuzzy_score(name1, name2[:len1], max_score - (len2 - len1), min_length)



import os
import shutil
import numpy as np
import copy
import tempfile
from .atom import Atom
from .molecule import Molecule
from .unitcell import UnitCell
from ..forcefield import ForceField


class Topology():
    '''
    A topology is a combination of molecules and unit cell.

    Parameters
    ----------
    molecules : list of Molecule, optional
        If molecules not provided, the topology should be initialized by calling update_molecules().
    numbers : list of int, optional
        The number of each molecule you want to put in this topology.
    cell : UnitCell, optional
        If cell not provided, a default cell with zero volume will be assumed.

    Notes
    -----
    All the molecules will be deep copied if there are identical molecules in the list, or if numbers is not None.
    The cell will always be deep copied if provided.

    Attributes
    ----------
    remark : str
        A short remark for this topology. Only used for output topology file
    cell : UnitCell
        Unit cell of this topology. Required for output simulation files and guessing connectivity
    '''

    _class_map = {}

    @staticmethod
    def register_format(extension, cls):
        Topology._class_map[extension] = cls

    def __init__(self, molecules=None, numbers=None, cell=None):
        self.remark = ''
        self._molecules: [Molecule] = []
        self._atoms: [Atom] = []
        if molecules is not None:
            self.update_molecules(molecules, numbers)
        if cell is not None:
            self.cell = UnitCell(cell.vectors)
        else:
            self.cell = UnitCell()

    def __repr__(self):
        return f'<Topology: {self.n_molecule} molecules {self.n_residue} residues {self.n_atom} atoms>'

    def __deepcopy__(self, memodict={}):
        top = Topology()
        top.remark = self.remark
        top.cell = UnitCell(self.cell.vectors)
        top.update_molecules(self._molecules, deepcopy=True)
        return top

    def update_molecules(self, molecules, numbers=None, deepcopy=False):
        '''
        Update the molecules contained in this topology.

        The index of molecules and atoms will be updated.
        Therefore, this method should be called as long as atoms or molecules are added to or removed from topology.

        Parameters
        ----------
        molecules : list of Molecule
        numbers : list of int, Optional
        deepcopy : bool

        Notes
        -----
        The molecules and atoms are deep copied if numbers is not None or deepcopy is True.
        The molecules and atoms are deep copied if there are identical molecules in the list.
        '''
        if not isinstance(molecules, (list, tuple)) \
                or any(type(mol) is not Molecule for mol in molecules):
            raise Exception('A list of molecules should be provided')

        if len(set(molecules)) < len(molecules) and numbers is None:
            numbers = [1] * len(molecules)

        if numbers is None and not deepcopy:
            self._molecules = molecules[:]
        else:
            if numbers is None:
                numbers = [1] * len(molecules)
            elif len(molecules) != len(numbers):
                raise Exception('Elements in molecules and numbers do not match')
            self._molecules = []
            for mol, number in zip(molecules, numbers):
                for i in range(number):
                    self._molecules.append(copy.deepcopy(mol))

        for mol in self._molecules:
            mol._topology = self

        # cache atoms
        self._atoms = [atom for mol in self._molecules for atom in mol.atoms]

        # assign the id of molecules, atoms and residues belonging to this topology
        for i, mol in enumerate(self._molecules):
            mol.id = i
        for i, atom in enumerate(self._atoms):
            atom.id = i
        for i, residue in enumerate(self.residues):
            residue.id = i

    def set_positions(self, positions):
        '''
        Set the positions of all atoms in this topology

        Parameters
        ----------
        positions : array_like
        '''
        if self.n_atom != len(positions):
            raise Exception('Length of positions should equal to the number of atoms')
        for i, atom in enumerate(self.atoms):
            atom.position = positions[i]

    def get_unique_molecules(self, deepcopy=True):
        '''
        Get the unique molecules and the number of molecules that are similar to these unique molecules.

        This is mainly used for exporting GROMACS topol file.
        By default, the unique molecules are deep copied.

        Parameters
        ----------
        deepcopy : bool

        Returns
        -------
        molecule_number : dict [Molecule, int]
        '''
        mols_unique = []
        flags = [-1] * self.n_molecule
        mol_last: Molecule = None
        for i, mol in enumerate(self._molecules):
            if mol_last is not None and mol.is_similar_to(mol_last):
                flags[mol.id] = flags[mol_last.id]
            else:
                mols_unique.append(mol)
                flags[mol.id] = mol.id
            mol_last = mol

        if deepcopy:
            return {copy.deepcopy(mol): flags.count(mol.id) for mol in mols_unique}
        else:
            return {mol: flags.count(mol.id) for mol in mols_unique}

    def scale_with_packmol(self, numbers, packmol=None, seed=1, tempdir=None):
        '''
        Enlarge the number of molecules in this topology with Packmol.

        It is used for building real simulation box from several molecules.
        It generates molecules with appropriate positions and adds them into the topology.
        The unit cell should already be defined, and it will not be touched during the scaling.

        An input file for packmol and xyz files for molecules will be created under `tempdir`.
        If packmol is provided, then it will be invoked to pack the simulation box,
        and the generated positions will be loaded back to the topology,
        and the `tempdir` will be removed.
        Otherwise, the topology will be enlarged without correct positions,
        and the files under `tempdir` will be kept so that positions can be built manually and loaded back to the topology later.

        Parameters
        ----------
        numbers : int or list of int
            If numbers is a int, the topology will be scaled by this times.
            If numbers is a list of int, then each molecule will be scaled by corresponding times.
        packmol : Packmol, Optional
            If is None, will generate input files for packmol.
        seed : int
            Seed for random number for Packmol
        tempdir : str, Optional
            The temporary directory to keep intermediate files for packmol. It should exist already
            If is None, will create a directory under system temporary directory
        '''
        if type(numbers) is int:
            numbers = [numbers] * self.n_molecule
        else:
            if len(numbers) != self.n_molecule:
                raise Exception('Length of numbers should equal to molecules')
            if any(type(n) is not int for n in numbers):
                raise Exception('Numbers should be list of integers')

        if self.cell.volume == 0:
            raise Exception('Unit cell required by Packmol')

        if not self.cell.is_rectangular:
            raise Exception('Triclinic unit cell haven\'t been implemented')

        if not self.has_position:
            raise Exception('Positions required by Packmol')

        from ..wrapper.packmol import Packmol
        if packmol is not None and not isinstance(packmol, Packmol):
            raise Exception('Invalid Packmol')

        tempdir = tempdir or tempfile.mkdtemp(prefix='mstk-packmol-')

        xyz_files = []
        for i, mol in enumerate(self._molecules):
            xyz = tempdir + '/_MO_%i.xyz' % i
            xyz_files.append(xyz)
            Topology([mol]).write(xyz)

        tmp_inp = tempdir + '/_pack.inp'
        tmp_out = tempdir + '/_out.xyz'
        if packmol is not None:
            packmol.build_box(xyz_files, numbers, tmp_out,
                              size=self.cell.size - 0.2, inp_file=tmp_inp, seed=seed, silent=True)
            try:
                xyz = Topology.open(tmp_out)
            except:
                raise Exception('Building failed. xyz not found')

            if xyz.n_atom != sum(mol.n_atom * n for mol, n in zip(self._molecules, numbers)):
                raise Exception('Building failed. incorrectly information in xyz')

            self.update_molecules(self._molecules, numbers)
            self.set_positions(xyz.positions)
            shutil.rmtree(tempdir)
        else:
            Packmol.gen_inp(xyz_files, numbers, tmp_out, size=self.cell.size - 0.2, inp_file=tmp_inp, seed=seed)

    def scale_box(self, scale, rigid_group='atom'):
        '''
        Scale the box and positions of atoms

        Parameters
        ----------
        scale : float or np.array of shape (3,)
        rigid_group : ['atom', 'molecule', 'residue']
            If set to atom, the positions of all atoms will be equally scaled
            If set to molecule, each molecule will be treated as a rigid body during scaling
            If set to residue, each residue will be treated as a rigid body during scaling
        '''
        self.cell.set_box(self.cell.vectors * scale)
        if rigid_group == 'atom':
            for atom in self.atoms:
                atom.position = atom.position * scale
                return None

        if rigid_group == 'molecule':
            groups = self.molecules
        elif rigid_group == 'residue':
            groups = self.residues
        else:
            raise Exception(f'Invalid rigid_group: {rigid_group}')

        positions = self.positions
        masses = np.array([a.mass for a in self.atoms])
        for group in groups:
            ids = [a.id for a in group.atoms]
            pos_com = np.sum(positions[ids] * masses[ids, np.newaxis], axis=0) / np.sum(masses[ids])
            for atom in group.atoms:
                atom.position = pos_com * scale + atom.position - pos_com

    @property
    def n_molecule(self):
        '''
        Number of molecules belong to this topology

        Returns
        -------
        n : int
        '''
        return len(self._molecules)

    @property
    def n_atom(self):
        '''
        Number of atoms belong to this topology

        Returns
        -------
        n : int
        '''
        return len(self._atoms)

    @property
    def n_residue(self):
        '''
        Number of residues belong to this topology

        Returns
        -------
        n : int

        '''
        return sum(mol.n_residue for mol in self._molecules)

    @property
    def molecules(self):
        '''
        List of molecules belong to this topology

        Returns
        -------
        molecules : list of Molecule
        '''
        return self._molecules

    @property
    def atoms(self):
        '''
        List of atoms belong to this topology

        Returns
        -------
        atoms: list of Atom
        '''
        return self._atoms

    @property
    def residues(self):
        '''
        List of residues belonging to this topology

        The residue list is generated on the fly because it is called irregularly.
        Therefore, the performance of this method is not of concern.

        Returns
        -------
        residues : list of Residue
        '''
        return [residue for mol in self._molecules for residue in mol.residues]

    @property
    def n_bond(self):
        '''
        Number of bonds belong to this topology

        Returns
        -------
        n : int
        '''
        return sum([mol.n_bond for mol in self._molecules])

    @property
    def n_angle(self):
        '''
        Number of angles belong to this topology

        Returns
        -------
        n : int
        '''
        return sum([mol.n_angle for mol in self._molecules])

    @property
    def n_dihedral(self):
        '''
        Number of dihedrals belong to this topology

        Returns
        -------
        n : int
        '''
        return sum([mol.n_dihedral for mol in self._molecules])

    @property
    def n_improper(self):
        '''
        Number of impropers belong to this topology

        Returns
        -------
        n : int
        '''
        return sum([mol.n_improper for mol in self._molecules])

    @property
    def bonds(self):
        '''
        List of bonds belong to this topology

        Returns
        -------
        bonds : list of Bond
        '''
        return [bond for mol in self._molecules for bond in mol.bonds]

    @property
    def angles(self):
        '''
        List of angles belong to this topology

        Returns
        -------
        angles : list of Angle
        '''
        return [angle for mol in self._molecules for angle in mol.angles]

    @property
    def dihedrals(self):
        '''
        List of dihedrals belong to this topology

        Returns
        -------
        dihedrals : list of Dihedral
        '''
        return [dihedral for mol in self._molecules for dihedral in mol.dihedrals]

    @property
    def impropers(self):
        '''
        List of impropers belong to this topology

        Returns
        -------
        impropers : list of Improper
        '''
        return [improper for mol in self._molecules for improper in mol.impropers]

    @property
    def has_position(self):
        '''
        Whether or not all the atoms in the topology have positions

        Returns
        -------
        has : bool
        '''
        return all(atom.has_position for atom in self.atoms)

    @property
    def positions(self):
        '''
        Positions of all the atoms in this topology

        Returns
        -------
        positions : array_like
        '''
        return np.array([atom.position for atom in self.atoms])

    @property
    def is_drude(self):
        '''
        Whether or not this topology represents a Drude model

        Returns
        -------
        is : bool
        '''
        return any(atom.is_drude for atom in self._atoms)

    @property
    def has_virtual_site(self):
        '''
        Whether or not this topology contains virtual site

        Returns
        -------
        has : bool
        '''
        return any(atom.virtual_site is not None for atom in self._atoms)

    @staticmethod
    def open(file, **kwargs):
        '''
        Load topology from a file or smiles string

        The format will be determined from the extension of the file.

        Parameters
        ----------
        file : str
            If start with ':', then it will be treated as SMLIES string.
            Otherwise, it is the file to be read.

        Returns
        -------
        topology : Topology
        '''
        if file.startswith(':'):
            mol = Molecule.from_smiles(file[1:])
            return Topology([mol])

        basename, ext = os.path.splitext(file)
        ext = ext.lower()
        try:
            cls = Topology._class_map[ext]
        except KeyError:
            raise Exception('Unsupported topology format: ' + file)

        return cls(file, **kwargs).topology

    def write(self, file, **kwargs):
        '''
        Write topology to a file.

        The format will be determined from the extension of the file.

        Parameters
        ----------
        file : str
        '''
        basename, ext = os.path.splitext(file)
        ext = ext.lower()
        try:
            cls = Topology._class_map[ext]
        except KeyError:
            raise Exception('Unsupported format for topology: ' + ext)

        cls.save_to(self, file, **kwargs)

    @staticmethod
    def save_to(top, file, **kwargs):
        '''
        This method should be implemented by subclasses
        '''
        raise NotImplementedError('This method haven\'t been implemented')

    def to_omm_topology(self):
        '''
        Generate OpenMM topology from this topology

        Returns
        -------
        omm_topology : mm.app.Topology
        '''
        try:
            from openmm import app as omm_app
            from openmm.app.element import Element as omm_Element
        except ImportError:
            raise Exception('cannot import openmm')

        omm_top = omm_app.Topology()
        omm_chain = omm_top.addChain()
        d_omm_residues = {}
        for residue in self.residues:
            omm_res = omm_top.addResidue(residue.name, omm_chain)
            for atom in residue.atoms:
                d_omm_residues[atom] = omm_res
        for atom in self.atoms:
            try:
                omm_element = omm_Element.getBySymbol(atom.symbol)
            except:
                omm_element = None
            omm_top.addAtom(atom.name, omm_element, d_omm_residues[atom])

        if self.cell.volume != 0:
            omm_top.setPeriodicBoxVectors(self.cell.vectors)
        omm_atoms = list(omm_top.atoms())
        for bond in self.bonds:
            omm_top.addBond(omm_atoms[bond.atom1.id], omm_atoms[bond.atom2.id])

        return omm_top

    def init_from_omm_topology(self, omm_top):
        '''
        Init the topology from OpenMM topology.

        This is only used for write trajectory, therefore only the atomic symbol is loaded.
        It should not be used for other purpose.

        Parameters
        ----------
        omm_top : mm.app.Topology
        '''
        from openmm.app import Topology as ommTopology, Residue as ommResidue, Atom as ommAtom
        molecules = []
        omm_top: ommTopology
        omm_res: ommResidue
        omm_atom: ommAtom
        for omm_res in omm_top.residues():
            mol = Molecule(omm_res.name)
            for omm_atom in omm_res.atoms():
                atom = Atom(omm_atom.name)
                try:
                    atom.symbol = omm_atom.element.symbol
                except:
                    pass
                mol.add_atom(atom)
            molecules.append(mol)
        self.update_molecules(molecules)

    def get_drude_pairs(self):
        '''
        Retrieve all the Drude dipole pairs belong to this topology

        Returns
        -------
        pairs :  list of tuple of Atom
            [(parent, drude)]

        See Also
        --------
        Molecule.get_drude_pairs
        '''
        return [pair for mol in self._molecules for pair in mol.get_drude_pairs()]

    def get_virtual_site_pairs(self):
        '''
        Retrieve all the virtual site pairs belong to this topology

        TODO It assumes no more than one virtual site is attached to each atom

        Returns
        -------
        pairs :  list of tuple of Atom
            [(parent, atom_virtual_site)]

        See Also
        --------
        Molecule.get_virtual_site_pairs
        '''
        return [pair for mol in self._molecules for pair in mol.get_virtual_site_pairs()]

    def get_12_13_14_pairs(self):
        '''
        Retrieve all the 1-2, 1-3 and 1-4 pairs based on the bond information.

        The pairs only concerns real atoms. Drude particles  will be ignored.

        Returns
        -------
        pairs12 : list of tuple of Atom
        pairs13 : list of tuple of Atom
        pairs14 : list of tuple of Atom

        See Also
        --------
        Molecule.get_12_13_14_pairs
        '''
        pair_12_list = []
        pair_13_list = []
        pair_14_list = []
        mol: Molecule
        for mol in self._molecules:
            pairs_12, pairs_13, pairs_14 = mol.get_12_13_14_pairs()
            pair_12_list += pairs_12
            pair_13_list += pairs_13
            pair_14_list += pairs_14
        return pair_12_list, pair_13_list, pair_14_list

    def generate_angle_dihedral_improper(self, dihedral=True, improper=True):
        '''
        Generate angles, dihedrals and impropers from bonds.

        See Also
        --------
        Molecule.generate_angle_dihedral_improper
        '''
        for mol in self._molecules:
            mol.generate_angle_dihedral_improper(dihedral=dihedral, improper=improper)

    def guess_connectivity_from_ff(self, ff, bond_limit=0.25, bond_tolerance=0.025, angle_tolerance=None, pbc=''):
        '''
        Guess the connectivity between atoms from force field parameters.

        Parameters
        ----------
        ff : ForceField
        bond_limit : float
        bond_tolerance : float
        angle_tolerance : float
        pbc : str

        See Also
        --------
        Molecule.guess_connectivity_from_ff
        '''
        mol: Molecule
        for mol in self._molecules:
            mol.guess_connectivity_from_ff(ff, bond_limit, bond_tolerance,
                                           angle_tolerance, pbc, self.cell)

    def generate_drude_particles(self, ff, **kwargs):
        '''
        Generate Drude particles from DrudeTerms in a polarizable force field

        Parameters
        ----------
        ff : ForceField
        kwargs : dict

        See Also
        --------
        Molecule.generate_drude_particles
        '''
        mol: Molecule
        for mol in self._molecules:
            mol.generate_drude_particles(ff, update_topology=False)
        self.update_molecules(self._molecules, deepcopy=False)

    def remove_drude_particles(self):
        '''
        Remove all the Drude particles from this topology

        It is useful for build non-polarizable model from polarizable model

        See Also
        --------
        Molecule.remove_drude_particles
        '''
        mol: Molecule
        for mol in self._molecules:
            mol.remove_drude_particles(update_topology=False)
        self.update_molecules(self._molecules, deepcopy=False)

    def generate_virtual_sites(self, ff, **kwargs):
        '''
        Generate virtual sites from VirtualSiteTerms in a force field

        Parameters
        ----------
        ff : ForceField
        kwargs : dict

        See Also
        --------
        Molecule.generate_virtual_sites
        '''
        mol: Molecule
        for mol in self._molecules:
            mol.generate_virtual_sites(ff, update_topology=False)
        self.update_molecules(self._molecules, deepcopy=False)

    def remove_virtual_sites(self):
        '''
        Remove all the virtual sites from this topology

        It is useful for build non-virtual-site model from virtual-site model

        See Also
        --------
        Molecule.remove_virtual_sites
        '''
        mol: Molecule
        for mol in self._molecules:
            mol.remove_virtual_sites(update_topology=False)
        self.update_molecules(self._molecules, deepcopy=False)



import datetime
import os
import time
from mstk import logger
from mstk.misc import DocstringMeta
from .pbsjob import PbsJob


class Scheduler(metaclass=DocstringMeta):
    '''
    Base class for job schedulers.

    Scheduler should not be constructed directly. Use its subclasses instead.

    Attributes
    ----------
    queue : str
        The jobs will be submitted to this queue.
    n_proc : int
        The CPU cores a job can use.
    n_gpu : int
        The GPU card a job can use.
    env_cmd : str, Optional
        The commands for setting up the environment before running real calculations.
    sh : str
        The default name of the job script
    max_running_hour: int
        The wall time limit for a job in hours.
    username : str
        The current user
    cached_jobs_expire : int
        The lifetime of cached jobs in seconds.
    '''

    #: Whether or not this is a remote job scheduler
    is_remote = False

    def __init__(self, queue=None, n_proc=1, n_gpu=0, env_cmd=None):
        self.queue = queue
        self.n_proc = n_proc
        self.n_gpu = n_gpu
        self.env_cmd = env_cmd or ''
        self.remote_dir = None

        if os.name == 'nt':
            import win32api
            self.username = win32api.GetUserName()
        else:
            import pwd
            self.username = pwd.getpwuid(os.getuid()).pw_name

        self.max_running_hour = 24
        self.sh = '_job.sh'
        self.cached_jobs_expire = 60  # seconds
        self._cached_jobs = []
        self._cache_last_update = None

    @property
    def all_jobs(self):
        '''
        Retrieve all the jobs that are currently managed by job scheduler.

        It call `scontrol show job` for Slurm or `qstat -f -u` for Torque to get the list of jobs.
        If some jobs have finished long max_running_hour ago (depends on the setting of job scheduler on the machine),
        they may disappeared from the list outputted by job scheduler.

        In order not to apply too much pressure to the job scheduler, cache is used to stores the jobs.
        This method will return the cached results without calling `scontrol` or `qstat` is the cache is not expired.
        The lifetime of cache is determined by attribute :attr:`cached_jobs_expire` in seconds.

        Returns
        -------
        jobs : list of PbsJob
        '''
        if self._cache_last_update is None or (
                datetime.datetime.now() - self._cache_last_update).total_seconds() >= self.cached_jobs_expire:
            self._update_stored_jobs()
        return self._cached_jobs

    def is_working(self):
        '''
        Whether or not this job scheduler is running normally.

        Returns
        -------
        is : bool
        '''
        raise NotImplementedError('This method should be implemented by subclasses')

    def _update_stored_jobs(self):
        logger.info('Update job information')
        self._cached_jobs = []
        jobs = []
        for i in [3, 2, 1]:
            # in case get_all_jobs() raise Exception
            try:
                jobs = self.get_all_jobs()
            except Exception as e:
                logger.warning(repr(e))
            if jobs:
                break
            # in case get_all_jobs() failed, retry after 1s
            if i > 1:
                time.sleep(1)
        for job in reversed(jobs):  # reverse the job list, in case jobs with same name
            if job not in self._cached_jobs:
                self._cached_jobs.append(job)

        self._cached_jobs.reverse()  # reverse the job list
        self._cache_last_update = datetime.datetime.now()

    def generate_sh(self, commands, name, workdir=None, sh=None):
        '''
        Generate a shell script for commands to be executed by the job scheduler on compute nodes.

        Parameters
        ----------
        commands : str
            List of commands to be executed by the job scheduler on compute node step by step.
        name : str
            The name of the job to be submitted.
        workdir : str, Optional
            The working directory.
        sh : str
            The name (path) of the shell script being written.
            If not set, will use the default :attr:`sh`.
        '''
        raise NotImplementedError('This method should be implemented by subclasses')

    def upload(self, **kwargs) -> bool:
        '''
        Upload the simulation files to target folder.

        This method should be implemented by subclasses that is remote scheduler,
        which is determined by the attribute :attr:`is_remote`.
        If it's not a remote job scheduler, will simply return True.

        Returns
        -------
        successful : bool
            Whether or not the upload is successful
        '''
        pass

    def download(self, **kwargs) -> bool:
        '''
        Download the simulation files to target folder.

        This method should be implemented by subclasses that is remote scheduler,
        which is determined by the attribute :attr:`is_remote`.
        If it's not a remote job scheduler, will simply return True.

        Returns
        -------
        successful : bool
            Whether or not the download is successful
        '''
        pass

    def submit(self, sh=None, **kwargs):
        '''
        Submit a job script to scheduler.

        Parameters
        ----------
        sh : str, optional
            The file name of the job script.
            If not set, will use the default :attr:`sh`.

        Returns
        -------
        id : int
            Job ID. -1 means failed
        '''
        raise NotImplementedError('This method should be implemented by subclasses')

    def get_job_from_name(self, name):
        '''
        Get the job with specified name.

        If such a job can not be found, None will be returned.
        If several job have same name, the most recently submitted one will be returned.

        Parameters
        ----------
        name : str

        Returns
        -------
        job : PbsJob or None
        '''
        # if several job have same name, return the one with the largest id (most recent job)
        for job in sorted(self.all_jobs, key=lambda x: x.id, reverse=True):
            if job.name == name:
                return job
        return None

    def is_running(self, name):
        '''
        Check whether or not a job is pending or running (not killed or finished or failed).

        Parameters
        ----------
        name : str

        Returns
        -------
        is : bool
        '''
        job = self.get_job_from_name(name)
        if job is None:
            return False
        return job.state in [PbsJob.State.PENDING, PbsJob.State.RUNNING]

    def kill_job(self, name):
        '''
        Kill a job which has the specified name.

        Parameters
        ----------
        name : str

        Returns
        -------
        killed : bool
        '''
        raise NotImplementedError('This method should be implemented by subclasses')

    def get_all_jobs(self):
        '''
        Retrieve all the jobs that are currently managed by job scheduler.

        It call `scontrol show job` for Slurm or `qstat -f -u` for Torque to get the list of jobs.
        If some jobs have finished long max_running_hour ago (depends on the setting of job scheduler on the machine),
        they may disappeared from the list outputted by job scheduler.

        The difference between this method and property :attr:`all_jobs` is that, this method does not use the cached job list.
        Therefore it get the up to date jobs, but it also gives more pressure to the job scheduler.
        Usually, just call :attr:`all_jobs` instead of this method.

        Returns
        -------
        jobs : list of PbsJob
        '''
        raise NotImplementedError('This method should be implemented by subclasses')

    @property
    def n_running_jobs(self) -> int:
        '''
        The number of jobs that is currently pending or running (not killed or finished or failed)

        Only the jobs belongs to the specified :attr:`queue` will be considered.

        Returns
        -------
        n : int
        '''
        self._update_stored_jobs()  # must update stored jobs first
        n = 0
        for job in self.all_jobs:
            if job.state != PbsJob.State.DONE and job.queue == self.queue:
                n += 1
        return n



import io
from mstk.topology.atom import Atom
from mstk.topology.molecule import Molecule
from mstk.topology.topology import Topology
from mstk.chem.element import Element


class Msd():
    '''
    Generate Topology from MSD file of DFF.

    The atom names, symbols, types, charges, positions, molecules, bonds and unit cell (if provided in the file) are parsed.
    The bond orders, formal charges and subsets are ignored.

    Parameters
    ----------
    file : str or file-like object
        MSD file
    kwargs : dict
        Ignored

    Attributes
    ----------
    topology : Topology

    Examples
    --------
    >>> msd = Msd('input.msd')
    >>> topology = msd.topology

    '''

    def __init__(self, file, **kwargs):
        if type(file) is str:
            with open(file) as f:
                content = f.read()
        elif isinstance(file, io.IOBase):
            content = file.read()
        else:
            raise Exception('A filename or a file object required')

        self.topology = Topology()
        self._parse(content)

    def _parse(self, content):
        lines = content.splitlines()
        # Ignore the first three lines
        # The forth line could be cell information or atom number
        line = lines[3]
        if line.startswith('PBC'):
            words = line.strip().split()
            lengths = list(map(lambda x: float(x) / 10, words[1:4]))
            angles = list(map(float, words[4:7]))
            self.topology.cell.set_box([lengths, angles])
            n_header = 4
        else:
            n_header = 3

        n_atom = int(lines[n_header])
        atoms = [Atom() for i in range(n_atom)]
        mol_names = {}
        for i in range(n_atom):
            line = lines[n_header + 1 + i]
            words = line.strip().split()
            atom_id = int(words[0])  # index in MSD starts from 1
            atom = atoms[atom_id - 1]
            atom.id = atom_id - 1  # atom.id stars from 0
            atom.name = words[1]
            element = Element(int(words[2]))
            atom.symbol = element.symbol
            atom.mass = element.mass
            atom.type = words[3]
            atom.charge, x, y, z = list(map(float, words[5:9]))
            atom.position = [x / 10, y / 10, z / 10]

            mol_id = int(words[9])  # starts from 1
            mol_name = words[10]
            if mol_id not in mol_names:
                mol_names[mol_id] = mol_name
            atom._mol_id = mol_id  # temporary attribute for identifying molecules

        molecules = [Molecule(name) for id, name in sorted(mol_names.items())]
        for atom in atoms:
            mol = molecules[atom._mol_id - 1]
            mol.add_atom(atom, update_topology=False)

        n_bond = int(lines[n_header + n_atom + 1])
        for i in range(n_bond):
            line = lines[n_header + n_atom + 2 + i]
            words = line.strip().split()
            atom1 = atoms[int(words[0]) - 1]
            atom2 = atoms[int(words[1]) - 1]
            atom1.molecule.add_bond(atom1, atom2)

        self.topology.update_molecules(molecules)
        self.topology.generate_angle_dihedral_improper()


Topology.register_format('.msd', Msd)



import math
from .. import GroFile

class GroReporter(object):
    '''
    GroReporter outputs a series of frames from a Simulation to a GRO file.

    Parameters
    ----------
    file : string
        The file to write to
    reportInterval : int
        The interval (in time steps) at which to write frames
    logarithm : bool
        If set to True, then write trajectory at logarithm interval.
        reportInterval will be the minimum step for reporting.
        e.g. when reportInterval set to 30, then report at [30, 40, 50, ..., 90, 100, 200, ..., 900, 1000, 2000, ...] steps.
    enforcePeriodicBox: bool
        Specifies whether particle positions should be translated
        so the center of every molecule lies in the same periodic box.
        If None (the default), it will automatically decide whether to translate molecules
        based on whether the system being simulated uses periodic boundary conditions.
    subset : list(int)=None
        If not None, only the selected atoms will be written
    reportVelocity: bool
        If set to True, velocities will be reported
    append: bool
        If set to True, will append to file
    '''

    def __init__(self, file, reportInterval, logarithm=False, enforcePeriodicBox=False, subset=None, reportVelocity=False, append=False):
        self._reportInterval = reportInterval
        self._logarithm = logarithm
        self._enforcePeriodicBox = enforcePeriodicBox
        if append:
            self._out = open(file, 'a')
        else:
            self._out = open(file, 'w')
        self._reportVelocity = reportVelocity

        if subset is None:
            self._subset = None
        else:
            self._subset = subset[:]

    def describeNextReport(self, simulation):
        """Get information about the next report this object will generate.

        Parameters
        ----------
        simulation : Simulation
            The Simulation to generate a report for

        Returns
        -------
        tuple
            A six element tuple. The first element is the number of steps
            until the next report. The next four elements specify whether
            that report will require positions, velocities, forces, and
            energies respectively.  The final element specifies whether
            positions should be wrapped to lie in a single periodic box.
        """
        if self._logarithm:
            if simulation.currentStep < self._reportInterval:
                _base = self._reportInterval
            else:
                _base = 10 ** math.floor(math.log10(simulation.currentStep))
            steps = _base - simulation.currentStep % _base
        else:
            steps = self._reportInterval - simulation.currentStep % self._reportInterval

        return (steps, True, self._reportVelocity, False, False, self._enforcePeriodicBox)

    def report(self, simulation, state):
        """Generate a report.

        Parameters
        ----------
        simulation : Simulation
            The Simulation to generate a report for
        state : State
            The current state of the simulation
        """
        time = state.getTime()
        positions = state.getPositions(asNumpy=True)
        velocities = state.getVelocities(asNumpy=True) if self._reportVelocity else None
        vectors = state.getPeriodicBoxVectors()
        GroFile.writeFile(simulation.topology, positions, vectors, self._out, time, self._subset, velocities)

        if hasattr(self._out, 'flush') and callable(self._out.flush):
            self._out.flush()

    def __del__(self):
        self._out.close()



import os
from io import IOBase


class TrjHandler():
    '''
    Base class of trajectory handlers for various trajectory formats.

    The methods :func:`get_info`, :func:`read_frame` and :func:`write_frame` should be implemented by subclasses.
    The method :func:`close` should also be overriden by subclasses if more works are required more than close the file.
    '''

    _klass_map = {}

    def __init__(self):
        self._file = IOBase()
        self.n_atom = -1
        self.n_frame = -1

    @staticmethod
    def register_format(extension, Handler):
        '''
        Register a handler class for a trajectory format based on the extension name

        Parameters
        ----------
        extension : str
            The extension should start with '.'.
            The upper of lower cases are not distinguished.
        Handler : subclass of TrjHandler
        '''
        TrjHandler._klass_map[extension.lower()] = Handler

    @staticmethod
    def get_handler_for_file(file):
        '''
        Get the appropriate handler class for a trajectory file.

        If a list of string is provided, the CombinedTrj will be returned.
        Otherwise, the handler class will be determined from the extension name of the file.

        Parameters
        ----------
        file : str or list of str
        '''
        from . import CombinedTrj

        if isinstance(file, list):
            Handler = CombinedTrj
        elif isinstance(file, str):
            try:
                Handler = TrjHandler._klass_map[os.path.splitext(file)[-1].lower()]
            except:
                raise Exception('Unknown extension for trajectory file: %s' % file)
        else:
            raise Exception('Only string and list of string are accepted for trajectory file')

        return Handler

    def get_info(self):
        '''
        Get the number of atoms and frames in the trajectory.

        Also record the offset of lines and frames, so that we can read arbitrary frame later.
        It assumes all frames have the same number of atoms.

        Returns
        -------
        n_atom : int
        n_frame : int
        '''
        raise NotImplementedError('Method not implemented')

    def read_frame(self, i_frame, frame):
        '''
        Read a single frame.

        Parameters
        ----------
        i_frame : int
            The index of the frame in the trajectory
        frame : Frame
            The information read from the trajectory will be written into this Frame
        '''
        raise NotImplementedError('Method not implemented')

    def write_frame(self, frame, **kwargs):
        '''
        Write a frame into the trajectory file opened by the handler.

        The handler should be initialized in mode 'w' or 'a'.

        Parameters
        ----------
        frame : Frame
        kwargs : dict
        '''
        raise NotImplementedError('Method not implemented')

    def close(self):
        '''
        Close the handler.
        '''
        self._file.close()



#!/usr/bin/env python3

import sys
import argparse

from mstk.forcefield import ForceField


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('input', nargs='+', type=str, help='force field files')
    parser.add_argument('-o', '--output', required=True, type=str, help='output file')
    return parser.parse_args()


if __name__ == '__main__':
    args = parse_args()

    ff = ForceField.open(*args.input)
    print('%6i atom types\n'
          '%6i virtual site terms\n'
          '%6i self vdW terms\n'
          '%6i pairwise vdW terms\n'
          '%6i charge increment terms\n'
          '%6i bond terms\n'
          '%6i angle terms\n'
          '%6i dihedral terms\n'
          '%6i improper terms\n'
          '%6i polarizable terms' % (
              len(ff.atom_types), len(ff.virtual_site_terms), len(ff.vdw_terms), len(ff.pairwise_vdw_terms),
              len(ff.qinc_terms), len(ff.bond_terms), len(ff.angle_terms),
              len(ff.dihedral_terms), len(ff.improper_terms), len(ff.polarizable_terms)
          ))

    ff.write(args.output)



import math
from mstk.chem.element import Element
from mstk.topology.atom import Atom
from mstk.topology.molecule import Molecule
from mstk.topology.topology import Topology
from mstk.topology.virtualsite import TIP4PSite


class Psf:
    '''
    Generate Topology from PSF file.

    Atoms, bonds, angles, dihedrals, impropers and virtual sites are parsed.
    Polarizability and thole parameters for Drude model are parsed,
    but anisotroic polarizability are ignored.
    All others like hydrogen bonds are ignored.

    Parameters
    ----------
    file : str
    split_molecule : str
        Can be 'residue', 'whole', 'bond' or 'auto'. Default is 'auto'.
        If set to 'residue', each residue will be parsed as a molecule. It works only if there is no inter-residue bonds.
        If set to 'whole', all the atoms will be put into one molecule.
        If set to 'bond', the atoms will be grouped into molecules based on connectivity.
        if set to 'auto', it will try 'residue' first. If there's inter-residue connectivity, will use 'whole' instead.

    Attributes
    ----------
    topology : Topology

    Examples
    --------
    >>> psf = Psf('input.psf')
    >>> topology = psf.topology

    >>> Psf.save_to(topology, 'output.psf')
    '''

    def __init__(self, file, **kwargs):
        self.topology = Topology()
        self._molecule = Molecule()  # a single molecule holds all the atoms
        self._parse(file, **kwargs)

    def _parse(self, file, split_molecule='auto'):
        with open(file) as f:
            lines = f.read().splitlines()

        words = lines[0].strip().split()
        if 'PSF' not in words:
            raise Exception('Invalid psf file, PSF not found in first line')
        is_drude = 'DRUDE' in words

        # there are other sections, but i don't care
        _sections = ['NTITLE', 'NATOM', 'NBOND', 'NTHETA', 'NPHI', 'NIMPHI', 'NUMLP NUMLPH']

        # record the line number (starts from 0) of section headers
        _iline_section = []
        for i, line in enumerate(lines):
            words = line.strip().split('!')
            if len(words) < 2:
                continue
            word = words[1].split(':')[0]
            if word in _sections:
                _iline_section.append((i, word))

        for i, (iline, section) in enumerate(_iline_section):
            if i != len(_iline_section) - 1:
                iline_next = _iline_section[i + 1][0]
            else:
                iline_next = len(lines)
            if section == 'NTITLE':
                # the number at section line indicates the line counts of the section
                self._parse_title(lines[iline: iline_next])
            if section == 'NATOM':
                self._parse_atoms(lines[iline: iline_next], is_drude)
            if section == 'NBOND':
                self._parse_bonds(lines[iline: iline_next])
            if section == 'NTHETA':
                self._parse_angles(lines[iline: iline_next])
            if section == 'NPHI':
                self._parse_dihedrals(lines[iline: iline_next])
            if section == 'NIMPHI':
                self._parse_impropers(lines[iline: iline_next])
            if section == 'NUMLP NUMLPH':
                self._parse_virtual_sites(lines[iline: iline_next])

        if split_molecule == 'residue':
            self.topology.update_molecules(self._molecule.split_residues())
        elif split_molecule == 'whole':
            self.topology.update_molecules([self._molecule])
        elif split_molecule == 'bond':
            self.topology.update_molecules(self._molecule.split(consecutive=True))
        elif split_molecule == 'auto':
            try:
                pieces = self._molecule.split_residues()
            except:
                self.topology.update_molecules([self._molecule])
            else:
                self.topology.update_molecules(pieces)
        else:
            raise Exception(f'Invalid option for split_molecule: {split_molecule}')

    def _parse_title(self, lines):
        nline = int(lines[0].strip().split()[0])
        self.topology.remark = '\n'.join(lines[1:1 + nline]).strip()

    def _parse_atoms(self, lines, parse_drude):
        '''
        in psf file, drude particles always appear after attached atoms
        '''
        mol = self._molecule
        n_atom = int(lines[0].strip().split()[0])
        res_names = {}  # {int: str}
        res_atoms = {}  # {int: [Atom]}
        for i in range(n_atom):
            words = lines[1 + i].strip().split()
            res_id = int(words[2])
            res_name = words[3]
            atom_name = words[4]
            atom_type = words[5]
            charge, mass = list(map(float, words[6:8]))
            if parse_drude:
                alpha, thole = list(map(float, words[9:11]))
            else:
                alpha = thole = 0.0

            atom = Atom(atom_name)
            atom.charge = charge
            atom.mass = mass
            atom.type = atom_type
            atom.alpha = -alpha / 1000  # convert A^3 to nm^3
            atom.thole = thole * 2
            if parse_drude and atom.type.startswith('D'):
                atom.is_drude = True
                atom.symbol = 'DP'
            else:
                atom.symbol = Element.guess_from_atom_type(atom.type).symbol

            mol.add_atom(atom)

            if res_id not in res_names:
                res_names[res_id] = res_name
                res_atoms[res_id] = []
            res_atoms[res_id].append(atom)

        for resid, resname in res_names.items():
            atoms = res_atoms[resid]
            mol.add_residue(resname, atoms, refresh_residues=False)
        mol.refresh_residues()

    def _parse_bonds(self, lines):
        n_bond = int(lines[0].strip().split()[0])
        atoms = self._molecule.atoms
        for i in range(math.ceil(n_bond / 4)):
            words = lines[i + 1].strip().split()
            for j in range(4):
                if i * 4 + j + 1 > n_bond:
                    break
                atom1, atom2 = atoms[int(words[j * 2]) - 1], atoms[int(words[j * 2 + 1]) - 1]
                atom1.molecule.add_bond(atom1, atom2)

    def _parse_angles(self, lines):
        n_angle = int(lines[0].strip().split()[0])
        atoms = self._molecule.atoms
        for i in range(math.ceil(n_angle / 3)):
            words = lines[i + 1].strip().split()
            for j in range(3):
                if i * 3 + j + 1 > n_angle:
                    break
                atom1, atom2, atom3 = atoms[int(words[j * 3 + 0]) - 1], \
                                      atoms[int(words[j * 3 + 1]) - 1], \
                                      atoms[int(words[j * 3 + 2]) - 1]
                atom1.molecule.add_angle(atom1, atom2, atom3)

    def _parse_dihedrals(self, lines):
        n_dihedral = int(lines[0].strip().split()[0])
        atoms = self._molecule.atoms
        for i in range(math.ceil(n_dihedral / 2)):
            words = lines[i + 1].strip().split()
            for j in range(2):
                if i * 2 + j + 1 > n_dihedral:
                    break
                atom1, atom2, atom3, atom4 = atoms[int(words[j * 4 + 0]) - 1], \
                                             atoms[int(words[j * 4 + 1]) - 1], \
                                             atoms[int(words[j * 4 + 2]) - 1], \
                                             atoms[int(words[j * 4 + 3]) - 1]
                atom1.molecule.add_dihedral(atom1, atom2, atom3, atom4)

    def _parse_impropers(self, lines):
        n_improper = int(lines[0].strip().split()[0])
        atoms = self._molecule.atoms
        for i in range(math.ceil(n_improper / 2)):
            words = lines[i + 1].strip().split()
            for j in range(2):
                if i * 2 + j + 1 > n_improper:
                    break
                atom1, atom2, atom3, atom4 = atoms[int(words[j * 4 + 0]) - 1], \
                                             atoms[int(words[j * 4 + 1]) - 1], \
                                             atoms[int(words[j * 4 + 2]) - 1], \
                                             atoms[int(words[j * 4 + 3]) - 1]
                atom1.molecule.add_improper(atom1, atom2, atom3, atom4)

    def _parse_virtual_sites(self, lines):
        '''
        Currently only TIP4P style virtual sites are parsed
        '''
        n_vsite, n_holder = map(int, lines[0].strip().split()[:2])
        if n_holder != 4 * n_vsite:
            raise Exception('Only TIP4P style virtual sites supported: ' + lines[0])

        d_list = [0] * n_vsite
        for i in range(n_vsite):
            words = lines[i + 1].strip().split()
            if len(words) != 6 or words[0] != '3' or words[2] != 'F':
                raise Exception('Only TIP4P style virtual sites supported: ' + lines[i + 1])
            d, theta, phi = map(float, words[-3:])
            if theta != 0 or phi != 0:
                raise Exception('Only TIP4P style virtual sites supported: ' + lines[i + 1])
            d_list[i] = -d / 10  # convert from A to nm

        atoms = self._molecule.atoms
        for i in range(math.ceil(n_vsite / 2)):
            words = lines[i + 1 + n_vsite].strip().split()
            for j in range(2):
                if i * 2 + j + 1 > n_vsite:
                    break
                atom1, atom2, atom3, atom4 = atoms[int(words[j * 4 + 0]) - 1], \
                                             atoms[int(words[j * 4 + 1]) - 1], \
                                             atoms[int(words[j * 4 + 2]) - 1], \
                                             atoms[int(words[j * 4 + 3]) - 1]
                atom1.virtual_site = TIP4PSite([atom2, atom3, atom4], [d_list[i * 2 + j]])

    @staticmethod
    def save_to(top, file, **kwargs):
        '''
        Save topology into a PSF file

        Parameters
        ----------
        top : Topology
        file : str
        '''
        for atom in top.atoms:
            if atom.type == '':
                raise Exception('PSF requires all atom types are defined')

        string = ''

        if top.is_drude:
            string += 'PSF DRUDE\n\n'
        else:
            string += 'PSF\n\n'

        string += '%8i !NTITLE\n' % 1
        string += ' REMARKS Created by mstk\n\n'

        string += '%8i !NATOM\n' % top.n_atom
        for i, atom in enumerate(top.atoms):
            string += '%8i %-4s %-4i %-4s %-4s %-4s %10.6f %13.4f %11i %8.4f %8.4f\n' % (
                atom.id + 1, 'S', atom.residue.id + 1, atom.residue.name, atom.name, atom.type,
                atom.charge, atom.mass, 0, -atom.alpha * 1000, atom.thole / 2
            )
        string += '\n'

        n_bond = top.n_bond
        string += '%8i !NBOND: bonds\n' % n_bond
        for i, bond in enumerate(top.bonds):
            string += '%8i%8i' % (bond.atom1.id + 1, bond.atom2.id + 1)
            if (i + 1) % 4 == 0 or i + 1 == n_bond:
                string += '\n'
        string += '\n'

        n_angle = top.n_angle
        string += '%8i !NTHETA: angles\n' % n_angle
        for i, angle in enumerate(top.angles):
            string += '%8i%8i%8i' % (angle.atom1.id + 1, angle.atom2.id + 1,
                                     angle.atom3.id + 1)
            if (i + 1) % 3 == 0 or i + 1 == n_angle:
                string += '\n'
        string += '\n'

        n_dihedral = top.n_dihedral
        string += '%8i !NPHI: dihedrals\n' % n_dihedral
        for i, dihedral in enumerate(top.dihedrals):
            string += '%8i%8i%8i%8i' % (dihedral.atom1.id + 1, dihedral.atom2.id + 1,
                                        dihedral.atom3.id + 1, dihedral.atom4.id + 1)
            if (i + 1) % 2 == 0 or i + 1 == n_dihedral:
                string += '\n'
        string += '\n'

        n_improper = top.n_improper
        string += '%8i !NIMPHI: impropers\n' % n_improper
        for i, improper in enumerate(top.impropers):
            string += '%8i%8i%8i%8i' % (improper.atom1.id + 1, improper.atom2.id + 1,
                                        improper.atom3.id + 1, improper.atom4.id + 1)
            if (i + 1) % 2 == 0 or i + 1 == n_improper:
                string += '\n'
        string += '\n'

        string += '%8i !NDON: donors\n\n' % 0
        string += '%8i !NACC: acceptors\n\n' % 0
        string += '%8i !NNB\n\n\n' % 0

        vsite_pairs = top.get_virtual_site_pairs()
        n_vsite = len(vsite_pairs)
        string += '%8i %i !NUMLP NUMLPH\n' % (n_vsite, 4 * n_vsite)
        for parent, avsite in vsite_pairs:
            if type(avsite.virtual_site) is not TIP4PSite:
                raise Exception('Virtual sites other than TIP4PSite haven\'t been implemented')
            string += '%8i%8i  F %10.6f %10.6f %10.6f\n' % (
                3, parent.id + 1, -avsite.virtual_site.parameters[0] * 10, 0, 0)
        for i, (parent, avsite) in enumerate(vsite_pairs):
            O, H1, H2 = avsite.virtual_site.parents
            string += '%8i%8i%8i%8i' % (avsite.id + 1, O.id + 1, H1.id + 1, H2.id + 1)
            if (i + 1) % 2 == 0 or i + 1 == n_vsite:
                string += '\n'
        string += '\n'

        string += '%8i !NUMANISO\n\n' % 0

        with open(file, 'wb') as f:
            f.write(string.encode())


Topology.register_format('.psf', Psf)



"""
statedatareporter.py: Outputs data about a simulation

This is part of the OpenMM molecular simulation toolkit originating from
Simbios, the NIH National Center for Physics-Based Simulation of
Biological Structures at Stanford, funded under the NIH Roadmap for
Medical Research, grant U54 GM072970. See https://simtk.org.

Portions copyright (c) 2012-2013 Stanford University and the Authors.
Authors: Peter Eastman
Contributors: Robert McGibbon, Zheng Gong

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS, CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from __future__ import absolute_import
from __future__ import print_function

__author__ = "Peter Eastman"
__version__ = "1.0"

try:
    import bz2

    have_bz2 = True
except:
    have_bz2 = False

try:
    import gzip

    have_gzip = True
except:
    have_gzip = False

import openmm.openmm as mm
import openmm.unit as unit
import math
import time
import numpy as np
from mstk.topology.geometry import find_clusters


class StateDataReporter(object):
    '''
    StateDataReporter outputs information about a simulation, such as energy and temperature, to a file.

    This reporter is modified from the StateDataReporter shipped with OpenMM python API.
    It adds support for reporting pressure, energy contribution of individual force group, collective variables and box size.

    To use it, create a StateDataReporter, then add it to the Simulation's list of reporters.  The set of
    data to write is configurable using boolean flags passed to the constructor.  By default the data is
    written in comma-separated-value (CSV) format, but you can specify a different separator to use.

    Parameters
    ----------
    file : string or file
        The file to write to, specified as a file name or file object
    reportInterval : int
        The interval (in time steps) at which to write frames
    step : bool=False
        Whether to write the current step index to the file
    time : bool=False
        Whether to write the current time to the file
    potentialEnergy : bool=False
        Whether to write the potential energy to the file
    kineticEnergy : bool=False
        Whether to write the kinetic energy to the file
    totalEnergy : bool=False
        Whether to write the total energy to the file
    temperature : bool=False
        Whether to write the instantaneous temperature to the file
    volume : bool=False
        Whether to write the periodic box volume to the file
    box: bool=True
        Whether to write the periodic box size to the file
    density : bool=False
        Whether to write the system density to the file
    progress : bool=False
        Whether to write current progress (percent completion) to the file.
        If this is True, you must also specify totalSteps.
    remainingTime : bool=False
        Whether to write an estimate of the remaining clock time until
        completion to the file.  If this is True, you must also specify
        totalSteps.
    speed : bool=False
        Whether to write an estimate of the simulation speed in ns/day to
        the file
    elapsedTime : bool=False
        Whether to write the elapsed time of the simulation in hours to
        the file.
    separator : string=','
        The separator to use between columns in the file
    systemMass : mass=None
        The total mass to use for the system when reporting density.  If
        this is None (the default), the system mass is computed by summing
        the masses of all particles.  This parameter is useful when the
        particle masses do not reflect their actual physical mass, such as
        when some particles have had their masses set to 0 to immobilize
        them.
    totalSteps : int=None
        The total number of steps that will be included in the simulation.
        This is required if either progress or remainingTime is set to True,
        and defines how many steps will indicate 100% completion.
    forceGroups : list=[]
        Write the energy contribution of each force group in this list separately
    cvs : list=[]
        Write the value of each collective variable in this list separately.
        Each element is a CustomCVForce object
    extra : dict={}
        Write extra data from a dictionary. The keys will be used as header names.
        The values can change during the simulation.
    '''

    def __init__(self, file, reportInterval, step=True, time=False, potentialEnergy=True,
                 kineticEnergy=False, totalEnergy=False, temperature=True, volume=False, box=True,
                 density=True, progress=False, remainingTime=False, speed=True, elapsedTime=False,
                 separator='\t', systemMass=None, totalSteps=None, append=False,
                 pressure=True, pxx=False, pyy=False, pzz=False, forceGroups=None, cvs=None, extra=None):
        self._reportInterval = reportInterval
        self._openedFile = isinstance(file, str)
        if (progress or remainingTime) and totalSteps is None:
            raise ValueError(
                'Reporting progress or remaining time requires total steps to be specified')
        if self._openedFile:
            if append:
                self._out = open(file, 'a')
            else:
                self._out = open(file, 'w')
        else:
            self._out = file
        self._step = step
        self._time = time
        self._potentialEnergy = potentialEnergy
        self._kineticEnergy = kineticEnergy
        self._totalEnergy = totalEnergy
        self._temperature = temperature
        self._volume = volume
        self._box = box
        self._density = density
        self._progress = progress
        self._remainingTime = remainingTime
        self._speed = speed
        self._elapsedTime = elapsedTime
        self._separator = separator
        self._totalMass = systemMass
        self._totalSteps = totalSteps
        self._hasInitialized = False
        self._needsPositions = pressure or pxx or pyy or pzz
        # self._needsVelocities = pxx or pyy or pzz
        self._needsVelocities = False
        self._needsForces = False
        self._needEnergy = potentialEnergy or kineticEnergy or totalEnergy or temperature \
                           or pressure or pxx or pyy or pzz

        self._boxSizeList = [[], [], []]

        self._pressure = pressure
        self._pxx = pxx
        self._pyy = pyy
        self._pzz = pzz
        self._forceGroups = forceGroups or []
        self._cvs = cvs or []
        self._extra = extra or {}

    def describeNextReport(self, simulation):
        """Get information about the next report this object will generate.

        Parameters
        ----------
        simulation : Simulation
            The Simulation to generate a report for

        Returns
        -------
        tuple
            A five element tuple. The first element is the number of steps
            until the next report. The remaining elements specify whether
            that report will require positions, velocities, forces, and
            energies respectively.
        """
        steps = self._reportInterval - simulation.currentStep % self._reportInterval
        return (steps, self._needsPositions, self._needsVelocities, self._needsForces, self._needEnergy, False)

    def report(self, simulation, state):
        """Generate a report.

        Parameters
        ----------
        simulation : Simulation
            The Simulation to generate a report for
        state : State
            The current state of the simulation
        """
        if not self._hasInitialized:
            self._initializeConstants(simulation)
            headers = self._constructHeaders(simulation)
            print('#"%s"' % ('"' + self._separator + '"').join(headers), file=self._out)
            try:
                self._out.flush()
            except AttributeError:
                pass
            self._initialClockTime = time.time()
            self._initialSimulationTime = state.getTime()
            self._initialSteps = simulation.currentStep
            self._hasInitialized = True

        # Check for errors.
        self._checkForErrors(simulation, state)

        # Query for the values
        values = self._constructReportValues(simulation, state)

        # Write the values.
        print(self._separator.join(str(v) for v in values), file=self._out)
        try:
            self._out.flush()
        except AttributeError:
            pass

    def _constructReportValues(self, simulation, state):
        """Query the simulation for the current state of our observables of interest.

        Parameters
        ----------
        simulation : Simulation
            The Simulation to generate a report for
        state : State
            The current state of the simulation

        Returns
        -------
        A list of values summarizing the current state of
        the simulation, to be printed or saved. Each element in the list
        corresponds to one of the columns in the resulting CSV file.
        """
        values = []
        box = state.getPeriodicBoxVectors()
        volume = box[0][0] * box[1][1] * box[2][2]
        clockTime = time.time()
        if self._progress:
            values.append('%.1f%%' % (100.0 * simulation.currentStep / self._totalSteps))
        if self._step:
            values.append(simulation.currentStep)
        if self._time:
            values.append('%.4f' % state.getTime().value_in_unit(unit.picosecond))
        if self._temperature:
            values.append('%.2f' % (2 * state.getKineticEnergy() / (
                    self._dof * unit.MOLAR_GAS_CONSTANT_R)).value_in_unit(unit.kelvin))
        if self._pressure:
            values.append(self._compute_pressure(simulation.context, state))
        if self._potentialEnergy:
            values.append(
                '%.4f' % state.getPotentialEnergy().value_in_unit(unit.kilojoules_per_mole))
        if self._kineticEnergy:
            values.append('%.4f' % state.getKineticEnergy().value_in_unit(unit.kilojoules_per_mole))
        if self._totalEnergy:
            values.append(
                '%.4f' % (state.getKineticEnergy() + state.getPotentialEnergy()).value_in_unit(
                    unit.kilojoules_per_mole))
        if self._volume:
            values.append('%.4f' % volume.value_in_unit(unit.nanometer ** 3))
        if self._box:
            values.append('%.4f' % box[0][0].value_in_unit(unit.nanometer))
            values.append('%.4f' % box[1][1].value_in_unit(unit.nanometer))
            values.append('%.4f' % box[2][2].value_in_unit(unit.nanometer))
        if self._density:
            values.append('%.4f' % (self._totalMass / volume).value_in_unit(
                unit.gram / unit.item / unit.milliliter))
        if self._speed:
            elapsedDays = (clockTime - self._initialClockTime) / 86400.0
            elapsedNs = (state.getTime() - self._initialSimulationTime).value_in_unit(
                unit.nanosecond)
            if elapsedDays > 0.0:
                values.append('%.3g' % (elapsedNs / elapsedDays))
            else:
                values.append('--')
        if self._elapsedTime:
            values.append('%.3g' % ((time.time() - self._initialClockTime) / 3600.0))
        if self._remainingTime:
            elapsedSeconds = clockTime - self._initialClockTime
            elapsedSteps = simulation.currentStep - self._initialSteps
            if elapsedSteps == 0:
                value = '--'
            else:
                estimatedTotalSeconds = (self._totalSteps - self._initialSteps) * elapsedSeconds / elapsedSteps
                remainingSeconds = int(estimatedTotalSeconds - elapsedSeconds)
                remainingDays = remainingSeconds // 86400
                remainingSeconds -= remainingDays * 86400
                remainingHours = remainingSeconds // 3600
                remainingSeconds -= remainingHours * 3600
                remainingMinutes = remainingSeconds // 60
                remainingSeconds -= remainingMinutes * 60
                if remainingDays > 0:
                    value = "%d:%d:%02d:%02d" % (remainingDays, remainingHours, remainingMinutes, remainingSeconds)
                elif remainingHours > 0:
                    value = "%d:%02d:%02d" % (remainingHours, remainingMinutes, remainingSeconds)
                elif remainingMinutes > 0:
                    value = "%d:%02d" % (remainingMinutes, remainingSeconds)
                else:
                    value = "0:%02d" % remainingSeconds
            values.append(value)

        bool_press = [self._pxx, self._pyy, self._pzz]
        if any(bool_press):
            values.extend(self._compute_anisotropic_pressure(simulation.context, state, *bool_press))

        for group in self._forceGroups:
            values.append(simulation.context.getState(getEnergy=True, groups={group}).getPotentialEnergy().
                          value_in_unit(unit.kilojoules_per_mole))
        for cv in self._cvs:
            values.append(cv.getCollectiveVariableValues(simulation.context)[0])
        if self._extra:
            values.extend(self._extra.values())

        self._boxSizeList[0].append(box[0][0].value_in_unit(unit.nanometer)),
        self._boxSizeList[1].append(box[1][1].value_in_unit(unit.nanometer)),
        self._boxSizeList[2].append(box[2][2].value_in_unit(unit.nanometer)),

        return values

    def _initializeConstants(self, simulation):
        """Initialize a set of constants required for the reports

        Parameters
        - simulation (Simulation) The simulation to generate a report for
        """
        system = simulation.system
        if self._temperature:
            # Compute the number of degrees of freedom.
            dof = 0
            for i in range(system.getNumParticles()):
                if system.getParticleMass(i) > 0 * unit.dalton:
                    dof += 3
            dof -= system.getNumConstraints()
            if any(type(system.getForce(i)) == mm.CMMotionRemover for i in
                   range(system.getNumForces())):
                dof -= 3
            self._dof = dof
        if self._density:
            if self._totalMass is None:
                # Compute the total system mass.
                self._totalMass = 0 * unit.dalton
                for i in range(system.getNumParticles()):
                    self._totalMass += system.getParticleMass(i)
            elif not unit.is_quantity(self._totalMass):
                self._totalMass = self._totalMass * unit.dalton

        if self._pressure or self._pxx or self._pyy or self._pzz:
            self._masses = np.array([simulation.system.getParticleMass(i).value_in_unit(unit.dalton)
                                     for i in range(simulation.system.getNumParticles())]) * unit.dalton
            try:
                self._constrained_groups = [list(ids) for ids in simulation.context.getConstrainedGroups()]
            except AttributeError:
                # getConstrainedGroups is only available in my forked version
                system: mm.System = simulation.system
                n_atom = system.getNumParticles()
                matrix = np.zeros([n_atom, n_atom], dtype=bool)
                for i in range(system.getNumConstraints()):
                    a1, a2, d = system.getConstraintParameters(i)
                    matrix[a1][a2] = True
                    matrix[a2][a1] = True
                self._constrained_groups = find_clusters(list(range(n_atom)), lambda x, y: matrix[x][y])

    def _constructHeaders(self, simulation):
        """Construct the headers for the CSV output

        Returns: a list of strings giving the title of each observable being reported on.
        """
        headers = []
        if self._progress:
            headers.append('Progress')
        if self._step:
            headers.append('Step')
        if self._time:
            headers.append('Time')
        if self._temperature:
            headers.append('Temp')
        if self._pressure:
            headers.append('Press')
        if self._potentialEnergy:
            headers.append('E_pot')
        if self._kineticEnergy:
            headers.append('E_kin')
        if self._totalEnergy:
            headers.append('E_tot')
        if self._volume:
            headers.append('Vol')
        if self._box:
            headers.append('Lx')
            headers.append('Ly')
            headers.append('Lz')
        if self._density:
            headers.append('Density')
        if self._speed:
            headers.append('Speed')
        if self._elapsedTime:
            headers.append('Elapsed')
        if self._remainingTime:
            headers.append('Remaining')
        if self._pxx:
            headers.append('Pxx')
        if self._pyy:
            headers.append('Pyy')
        if self._pzz:
            headers.append('Pzz')
        for group in self._forceGroups:
            names = set(force.getName() for force in simulation.system.getForces() if force.getForceGroup() == group)
            headers.append(f'E_{group}_{"+".join(names)}')
        for i, cv in enumerate(self._cvs):
            headers.append(f'CV_{i}')
        if self._extra:
            headers.extend(self._extra.keys())
        return headers

    def _checkForErrors(self, simulation, state):
        """Check for errors in the current state of the simulation

        Parameters
         - simulation (Simulation) The Simulation to generate a report for
         - state (State) The current state of the simulation
        """
        if self._needEnergy:
            energy = (state.getKineticEnergy() + state.getPotentialEnergy()).value_in_unit(
                unit.kilojoules_per_mole)
            if math.isnan(energy):
                raise ValueError('Energy is NaN')
            if math.isinf(energy):
                raise ValueError('Energy is infinite')

    def _compute_pressure(self, context: mm.Context, state: mm.State):
        '''
        Compute the isotropic pressure of a rectangular system
        '''
        box = state.getPeriodicBoxVectors(asNumpy=True)
        positions = state.getPositions(asNumpy=True)
        a, b, c = box
        volume = a[0] * b[1] * c[2]
        p_kinetic = (2 * state.getKineticEnergy() * unit.item / 3 / volume).value_in_unit(unit.bar)

        # Because the contribution of constraints to the pressure cannot be evaluated,
        # here I approximate it by removing the kinetic contribution of the internal motion of each constrained group.
        # TODO I'm not confident about this formula
        p_kinetic *= len(self._constrained_groups) * 3 / self._dof

        scale = 0.0001
        context.setPeriodicBoxVectors(a * (1 + scale), b * (1 + scale), c * (1 + scale))
        context.setPositions(positions * (1 + scale))
        context.applyConstraints(1E-6)
        U1 = context.getState(getEnergy=True).getPotentialEnergy()

        context.setPeriodicBoxVectors(a * (1 - scale), b * (1 - scale), c * (1 - scale))
        context.setPositions(positions * (1 - scale))
        context.applyConstraints(1E-6)
        U_1 = context.getState(getEnergy=True).getPotentialEnergy()

        dU = (U1 - U_1) * unit.item
        dV = ((1 + scale) ** 3 - (1 - scale) ** 3) * volume
        p_virial = -(dU / dV).value_in_unit(unit.bar)

        context.setPeriodicBoxVectors(*box)
        context.setPositions(positions)

        return p_kinetic + p_virial

    def _compute_anisotropic_pressure(self, context, state, pxx, pyy, pzz):
        '''
        Compute the anisotropic pressure of a rectangular system

        Parameters
        ----------
        context : mm.Context
        state : mm.State
        pxx : bool
            Whether or not the Pxx be calculated
        pyy : bool
            Whether or not the Pyy be calculated
        pzz : bool
            Whether or not the Pzz be calculated
        '''
        box = state.getPeriodicBoxVectors(asNumpy=True)
        positions = state.getPositions(asNumpy=True)
        volume = box[0][0] * box[1][1] * box[2][2]

        # TODO Assume kinetic energies are well-partitioned to each DOF
        p_kinetic = (2 * state.getKineticEnergy() * unit.item / 3 / volume).value_in_unit(unit.bar)

        # Because the contribution of constraints to the pressure cannot be evaluated,
        # here I approximate it by removing the kinetic contribution of the internal motion of each constrained group.
        # TODO I'm not confident about this formula
        p_kinetic *= len(self._constrained_groups) * 3 / self._dof

        scale = 0.0001
        pressures = []
        for index, _bool in enumerate([pxx, pyy, pzz]):
            if not _bool:
                continue
            scale_array = np.array([1.0, 1.0, 1.0])
            scale_array[index] = 1 + scale

            box_scaled = box._value.copy()
            box_scaled[index][index] = box_scaled[index][index] * (1 + scale)

            context.setPeriodicBoxVectors(*box_scaled)
            context.setPositions(positions * scale_array)
            context.applyConstraints(1E-6)
            U1 = context.getState(getEnergy=True).getPotentialEnergy()

            scale_array[index] = 1 - scale
            box_scaled[index][index] = box_scaled[index][index] / (1 + scale) * (1 - scale)

            context.setPeriodicBoxVectors(*box_scaled)
            context.setPositions(positions * scale_array)
            context.applyConstraints(1E-6)
            U_1 = context.getState(getEnergy=True).getPotentialEnergy()

            dU = (U1 - U_1) * unit.item
            dV = 2 * scale * volume
            p_virial = -(dU / dV).value_in_unit(unit.bar)

            pressures.append(p_kinetic + p_virial)

        context.setPeriodicBoxVectors(*box)
        context.setPositions(positions)

        return pressures

    def __del__(self):
        if self._openedFile:
            self._out.close()

    def getBoxSizeAverage(self, timeFraction=0.5):
        '''
        Get the averaged box size of the simulated system since this reporter was added.

        Parameters
        ----------
        timeFraction : float
            The fraction of data (in the end) used to calculate the average.
            e.g. If set to 0.5, then the last half of the box size data will be used for average.
            If larger than 1, then so many data points at the end will be used for average.

        Returns
        -------
        lx : float
        ly : float
        lz : float
        '''
        n = math.ceil(timeFraction) if timeFraction > 1 else math.ceil(len(self._boxSizeList[0]) * timeFraction)
        lx = sum(self._boxSizeList[0][-n:]) / n
        ly = sum(self._boxSizeList[1][-n:]) / n
        lz = sum(self._boxSizeList[2][-n:]) / n

        return (lx, ly, lz)



import math
from distutils.util import strtobool
from mstk import logger
from mstk.chem.constant import *
from .errors import *


class FFTermFactory:
    '''
    Factory class for force field terms.

    Mainly used for loading a FFTerm from xml element stored in ZFP file.
    '''

    _class_map = {}

    @staticmethod
    def register(klass, extra_names=None):
        '''
        Register a force field term so that it can be read from or write into a ZFP or ZFF file.

        Parameters
        ----------
        klass : subclass of FFTerm
        extra_names : list of str
        '''
        FFTermFactory._class_map[klass.__name__] = klass
        FFTermFactory._class_map[klass.get_alias()] = klass  # for short alias in ZFF
        extra_names = extra_names or []
        for name in extra_names:
            FFTermFactory._class_map[name] = klass

    @staticmethod
    def create_from_zfp(name, d):
        '''
        Reconstruct a FFTerm using a dict of attributes read from ZFP xml file.

        Essentially it finds the class for the term, and calls the constructor of the class.
        Every class to be reconstructed should have a class attribute `_zfp_attrs`.
        This dict records which attributes should be saved into ZFP xml file,
        and it is also used as the arguments for initializing a FFTerm.
        It also tells the data type of these attributes.

        Some force field terms can not be fully built from the constructor.
        Then :func:`FFTerm._from_zfp_extra` should be implemented for those terms,
        and it will be called to do some extra works.

        Parameters
        ----------
        name : str
            Class name of the term to be constructed
        d : dict, [str, str]
            Attributes loaded from ZFP xml element

        Returns
        -------
        term : subclass of FFTerm

        '''
        # must use kwargs instead of args to init the cls,
        # in case there is mistake in the sequence of init arguments
        try:
            cls = FFTermFactory._class_map[name]
        except:
            raise Exception('Invalid force field term: %s' % name)
        kwargs = {}
        adjustables = []
        for attr, func in cls._zfp_attrs.items():
            str_val = d[attr]
            if str_val.endswith('?'):
                str_val = str_val[:-1]
                if func is float:
                    adjustables.append(attr)
            val = func(str_val)
            if attr in cls._zfp_convert:
                val = cls._zfp_convert[attr][0](val)
            kwargs[attr] = val
        try:
            term = cls(**kwargs)
        except:
            raise Exception('Cannot init %s with kwargs %s' % (name, str(d)))
        term.adjustables = adjustables
        term._from_zfp_extra(d)
        return term

    @staticmethod
    def create_from_zff(line):
        '''
        Reconstruct a FFTerm using a line read from ZFF text file.

        Essentially it finds the class for the term, and calls the constructor of the class.
        Every class to be reconstructed should have a class attribute `_zfp_attrs`.
        This dict records which attributes should be saved into ZFP xml file,
        and it is also used as the arguments for initializing a FFTerm.
        It also tell the data type of these attributes.

        Some force field terms can not be fully built from the constructor.
        Then :func:`FFTerm._from_zff_extra` should be implemented for those terms,
        and it will be called to do some extra works.

        Parameters
        ----------
        line : str
            A line read from ZFF file

        Returns
        -------
        term : subclass of FFTerm

        '''
        # must use kwargs instead of args to init the cls,
        # in case there is mistake in the sequence of init arguments
        words = line.strip().split()
        name = words[0]
        try:
            cls = FFTermFactory._class_map[name]
        except KeyError:
            raise Exception('Invalid force field term: %s' % name)
        kwargs = {}
        adjustables = []
        for (attr, func), str_val in zip(cls._zfp_attrs.items(), words[1:]):
            if str_val.endswith('?'):
                str_val = str_val[:-1]
                if func is float:
                    adjustables.append(attr)
            val = func(str_val)
            if attr in cls._zfp_convert:
                val = cls._zfp_convert[attr][0](val)
            kwargs[attr] = val
        try:
            term = cls(**kwargs)
        except:
            raise Exception('Cannot init %s with line: %s' % (name, line))
        term.adjustables = adjustables
        term._from_zff_extra(line)

        return term


class FFTerm:
    '''
    Base class for all force field terms like :class:`AtomType`, :class:`BondTerm`, etc...

    Attributes
    ----------
    version : str
        Version is used mainly in :class:`PPF` format by DFF.
    comments : list of str
        Comments is useful for debugging when generating input files for simulation engines.
    '''

    _zfp_attrs = {}
    '''
    The attributes that will be saved into or loaded from ZFP or ZFF file.
    The key is the attribute name. The value is a function to parse the string to object.
    This dict must be overridden by every subclasses that can be saved into or reconstructed from ZFP or ZFF file.
    '''

    _zfp_convert = {}
    '''
    The parameters that will be converted when read/write parameters from/to ZFP or ZFF file.
    The key is the attribute name. The value is a tuple of two functions.
    The first one convert data from external to internal. The second one do the opposite.
    For example, the angle is in unit of radian internally. But it is in unit of degree in ZFP or ZFF file.
    Therefore, for AngleTerm, _zfp_convert have two functions for attribute 'theta'.
    The first one convert degree to radian, and the second one convert radian to degree.
    This dict must be overridden by subclasses that have different value internally and in ZFP or ZFF file.
    '''

    @classmethod
    def get_alias(cls):
        '''
        Return a short alias for the name of this class

        It is used to keep the line more compact in ZFF file
        '''
        return cls.__name__.replace('Term', '')

    def __init__(self):
        self.version = None
        self.comments: [str] = []
        self.adjustables: [str] = []  # parameters that are subject to optimization

    def __repr__(self):
        return '<%s: %s>' % (self.__class__.__name__, self.name)

    @property
    def name(self):
        '''
        The name of this force field term.

        Returns
        -------
        name : str
        '''
        raise NotImplementedError('This property should be implemented by subclasses')

    def to_zfp(self):
        '''
        Pack the attributes of a term into a dict so that can be saved into ZFP xml file.

        Every class to be packed should have a class property `zfp_attrs`.
        This dict records which attributes should be saved into ZFP xml file,
        and it is also used as the arguments for initializing a FFTerm.
        It also tell the data type of these attributes.

        Returns
        -------
        d : dict, [str, str]
        '''
        d = {}
        for attr, func in self._zfp_attrs.items():
            val = getattr(self, attr)
            if attr in self._zfp_convert:
                val = self._zfp_convert[attr][1](val)
            d[attr] = str(val)
            if func is float and attr in self.adjustables:
                d[attr] += '?'
        d.update(self._to_zfp_extra())
        return d

    def _to_zfp_extra(self):
        '''
        Pack extra information for some special terms.

        e.g. PeriodicDihedralTerm can contain arbitrary numbers of dihedral parameters.

        Returns
        -------
        d : dict, [str, str]
        '''
        return {}

    def _from_zfp_extra(self, d):
        '''
        Some force field terms can not be fully built from the constructor.
        This function is called by :func:`FFTermFactory.create_from_zfp` to do some extra works.

        e.g. PeriodicDihedralTerm need to rebuild its own storage of multiple dihedral parameters

        Parameters
        ----------
        d : dict, [str, str]
        '''
        pass

    def to_zff(self):
        '''
        Pack the attributes of a term into a string so that can be saved into a line in ZFF file.

        Every class to be packed should have a class property `zfp_attrs`.
        This dict records which attributes should be saved into ZFP xml file,
        and it is also used as the arguments for initializing a FFTerm.
        It also tells the data type of these attributes.

        Returns
        -------
        line : string
        '''
        line = '%-16s' % self.__class__.get_alias()
        for attr, func in self._zfp_attrs.items():
            val = getattr(self, attr)
            if attr in self._zfp_convert:
                val = self._zfp_convert[attr][1](val)
            if func is float:
                if val > 1000:
                    string = ' %8.1f' % val
                else:
                    string = ' %8.4f' % val
                string += '?' if attr in self.adjustables else ' '
            elif func is int:
                string = ' %2i' % val
            else:
                string = ' %-9s' % val
            line += string

        return line.rstrip()

    def to_zff_header(self):
        '''
        Header string to explain a line in ZFF format

        Returns
        -------
        header : str
        '''
        line = '#%-15s' % self.__class__.get_alias()
        for attr, func in self._zfp_attrs.items():
            if func is float:
                string = ' %8s ' % attr
            elif func is int:
                string = ' %2i' % attr
            else:
                string = ' %-9s' % attr
            line += string
        return line.rstrip()

    def _from_zff_extra(self, line):
        '''
        Some force field terms can not be fully built from the constructor.
        This function is called by :func:`FFTermFactory.create_from_zff` to do some extra works.

        e.g. PeriodicDihedralTerm need to rebuild its own storage of multiple dihedral parameters

        Parameters
        ----------
        line : str
        '''
        pass

    def evaluate_energy(self, val):
        '''
        Evaluate the energy for a force field term like HarmonicBondTerm, PeriodicDihedralTerm, etc...

        It is mainly for debugging. It is not (and can not be) implemented for all terms.

        Parameters
        ----------
        val : float
            The value of distance, angle, dihedral or improper.

        Returns
        -------
        energy : float
        '''
        raise NotImplementedError('This method is invalid for this term')


class AtomType(FFTerm):
    '''
    AtomType is the most fundamental element in a force field.
    It determines which :class:`VdwTerm`, :class:`BondTerm`, etc...
    will be used to describe the energy functions between a specific set of atoms.

    Equivalent table (EQT) is extensively used in `mstk`.
    It helps decreasing the number of force field terms significantly without losing generality and accuracy,
    thus making the force field extendable.

    Parameters
    ----------
    name : str
        The name of this atom type.
    mass : float , optional
        The mass of this atom type.
    charge : float , optional
        The charge of this atom type.
    eqt_q_inc : str , optional
        The equivalent type for bond charge increment parameters
    eqt_vdw : str , optional
        The equivalent type for vdW parameters
    eqt_bond : str , optional
        The equivalent type for bond parameters
    eqt_ang_c : str , optional
        The equivalent type for angle parameters if this atom type is the center of an angle
    eqt_ang_s: str , optional
        The equivalent type for angle parameters if this atom type is the side of an angle
    eqt_dih_c : str , optional
        The equivalent type for dihedral parameters if this atom type is the center of a dihedral
    eqt_dih_s: str , optional
        The equivalent type for dihedral parameters if this atom type is the side of a dihedral
    eqt_imp_c : str , optional
        The equivalent type for improper parameters if this atom type is the center of an improper
    eqt_imp_s : str , optional
        The equivalent type for improper parameters if this atom type is the side of an improper
    eqt_polar : str , optional
        The equivalent type for polarization parameters

    Attributes
    ----------
    mass : float
        The mass of this atom type. -1 means not provided in the force field.
        Mass is not mandatory, because finally we are going to take the masses from the :class:`~mstk.topology.Topology`.
        However, masses in force field can be used to assign the masses in topology.
    charge : float
        The charge of this atom type.
        Similar to mass, the charges we are finally going to use are from the :class:`~mstk.topology.Topology`.
        But charges in force field can be used to assign the charges in topology.
    eqt_q_inc : str
        The equivalent type for bond charge increment parameters
    eqt_vdw : str
        The equivalent type for vdW parameters
    eqt_bond : str
        The equivalent type for bond parameters
    eqt_ang_c : str
        The equivalent type for angle parameters if this atom type is the center of an angle
    eqt_ang_s: str
        The equivalent type for angle parameters if this atom type is the side of an angle
    eqt_dih_c : str
        The equivalent type for dihedral parameters if this atom type is the center of a dihedral
    eqt_dih_s: str
        The equivalent type for dihedral parameters if this atom type is the side of a dihedral
    eqt_imp_c : str
        The equivalent type for improper parameters if this atom type is the center of an improper
    eqt_imp_s : str
        The equivalent type for improper parameters if this atom type is the side of an improper
    eqt_polar : str
        The equivalent type for polarization parameters
    '''
    _zfp_attrs = {
        'name'     : str,
        'mass'     : float,
        'charge'   : float,
        'eqt_q_inc': str,
        'eqt_vdw'  : str,
        'eqt_bond' : str,
        'eqt_ang_c': str,
        'eqt_ang_s': str,
        'eqt_dih_c': str,
        'eqt_dih_s': str,
        'eqt_imp_c': str,
        'eqt_imp_s': str,
        'eqt_polar': str,
    }

    def __init__(self, name, mass=-1, charge=0,
                 eqt_q_inc=None, eqt_vdw=None, eqt_bond=None,
                 eqt_ang_c=None, eqt_ang_s=None, eqt_dih_c=None, eqt_dih_s=None,
                 eqt_imp_c=None, eqt_imp_s=None, eqt_polar=None):
        super().__init__()
        self._name = name
        # mass is only used for assign mass of topology in case mass not exist in topology
        # mass = -1 means unknown
        self.mass = mass
        self.charge = charge
        self.eqt_q_inc = eqt_q_inc or name
        self.eqt_vdw = eqt_vdw or name
        self.eqt_bond = eqt_bond or name
        self.eqt_ang_c = eqt_ang_c or name
        self.eqt_ang_s = eqt_ang_s or name
        self.eqt_dih_c = eqt_dih_c or name
        self.eqt_dih_s = eqt_dih_s or name
        self.eqt_imp_c = eqt_imp_c or name
        self.eqt_imp_s = eqt_imp_s or name
        self.eqt_polar = eqt_polar or name

    @property
    def name(self):
        '''
        The name of this atom type.

        :setter: Set the name of this atom type

        Returns
        -------
        name : str
        '''
        return self._name

    @name.setter
    def name(self, value):
        # this is a trick allowing from_zfp() to work correctly
        self._name = value

    @property
    def eqt_types(self):
        return [self.eqt_q_inc,
                self.eqt_vdw,
                self.eqt_bond,
                self.eqt_ang_c,
                self.eqt_ang_s,
                self.eqt_dih_c,
                self.eqt_dih_s,
                self.eqt_imp_c,
                self.eqt_imp_s,
                self.eqt_polar]

    def __lt__(self, other):
        return self.name < other.name

    def __gt__(self, other):
        return self.name > other.name


FFTermFactory.register(AtomType)


class ChargeIncrementTerm(FFTerm):
    '''
    ChargeIncrementTerm describes the offset of charge between two atom types when they are bonded.

    Two atom types and the offset from the second to first should be provided to construct this term.
    e.g. The following example means when h_1 and c_4 are bonded,
    0.06 element charges will be transferred from c_4 to h_1.

    >>> term = ChargeIncrementTerm('h_1', 'c_4', 0.06)

    During the initialization, the two atom types will be sorted by their string.
    The value will be modified to represent the offset from the second to the first atom types after sorting.

    >>> print(term.type1, term.type2, term.value)
        c_4 h_1 -0.06

    Parameters
    ----------
    type1 : str
        The string of first atom type
    type2 : str
        The string of second atom type
    value : float
        The offset from type2 to type1

    Attributes
    ----------
    type1 : str
        The string of first atom type
    type2 : str
        The string of second atom type
    value : float
        The offset from type2 to type1
    '''
    _zfp_attrs = {
        'type1': str,
        'type2': str,
        'value': float,
    }

    def __init__(self, type1, type2, value):
        super().__init__()
        if type1 == type2 and value != 0:
            raise ChargeIncrementNonZeroError('Non-zero charge increment between atoms of same type')

        at1, at2 = sorted([type1, type2])
        self.type1 = at1
        self.type2 = at2
        self.value = value if at1 == type1 else -value

    @property
    def name(self):
        return '%s,%s' % (self.type1, self.type2)

    def __lt__(self, other):
        return [self.type1, self.type2] < [other.type1, other.type2]

    def __gt__(self, other):
        return [self.type1, self.type2] > [other.type1, other.type2]


FFTermFactory.register(ChargeIncrementTerm)


class VdwTerm(FFTerm):
    '''
    Base class for all vdW terms.

    Subclasses should be implemented for each functional form,
    e.g. :class:`LJ126Term`, :class:`MieTerm`.

    During the initialization, the two atom types will be sorted by their string.

    Parameters
    ----------
    type1 : str
    type2 : str

    Attributes
    ----------
    type1 : str
    type2 : str
    '''

    def __init__(self, type1, type2):
        super().__init__()
        at1, at2 = sorted([type1, type2])
        self.type1 = at1
        self.type2 = at2

    @property
    def name(self):
        return '%s,%s' % (self.type1, self.type2)

    def __lt__(self, other):
        return [self.type1, self.type2] < [other.type1, other.type2]

    def __gt__(self, other):
        return [self.type1, self.type2] > [other.type1, other.type2]


class BondTerm(FFTerm):
    '''
    Base class for all bond terms.

    Subclasses should be implemented for each functional form,
    e.g. :class:`HarmonicBondTerm`.

    During the initialization, the two atom types will be sorted by their string.

    Parameters
    ----------
    type1 : str
    type2 : str
    length : float
    fixed : bool

    Attributes
    ----------
    type1 : str
    type2 : str
    length : float
    fixed : bool
    '''

    def __init__(self, type1, type2, length, fixed=False):
        super().__init__()
        at1, at2 = sorted([type1, type2])
        self.type1 = at1
        self.type2 = at2
        self.length = length
        self.fixed = fixed

    @property
    def name(self):
        return '%s,%s' % (self.type1, self.type2)

    def __lt__(self, other):
        return [self.type1, self.type2] < [other.type1, other.type2]

    def __gt__(self, other):
        return [self.type1, self.type2] > [other.type1, other.type2]


class AngleTerm(FFTerm):
    '''
    Base class for all angle terms.

    Subclasses should be implemented for each functional form,
    e.g. :class:`HarmonicAngleTerm`, :class:`SDKAngleTerm`.

    During the initialization, the two side atom types will be sorted by their string.

    Parameters
    ----------
    type1 : str
    type2 : str
    type3 : str
    theta : float
    fixed : bool

    Attributes
    ----------
    type1 : str
    type2 : str
    type3 : str
    theta : float
    fixed : bool
    '''

    # the angle in ZFP or ZFF file is in unit of degree. one digit is enough
    _zfp_convert = {
        'theta': (lambda x: x * DEG2RAD, lambda x: round(x * RAD2DEG, 2))
    }

    def __init__(self, type1, type2, type3, theta, fixed=False):
        super().__init__()
        at1, at3 = sorted([type1, type3])
        self.type1 = at1
        self.type2 = type2
        self.type3 = at3
        self.theta = theta
        self.fixed = fixed

    @property
    def name(self):
        return '%s,%s,%s' % (self.type1, self.type2, self.type3)

    def __lt__(self, other):
        return [self.type1, self.type2, self.type3] < [other.type1, other.type2, other.type3]

    def __gt__(self, other):
        return [self.type1, self.type2, self.type3] > [other.type1, other.type2, other.type3]

    @property
    def is_linear(self):
        '''
        Whether or not this angle term is linear.

        Returns
        -------
        is :  bool
        '''
        return False


class DihedralTerm(FFTerm):
    '''
    Base class for all dihedral terms.

    Subclasses should be implemented for each functional form,
    e.g. :class:`PeriodicDihedralTerm`.

    During the initialization, the sequence of atom types to use `i-j-k-l` or `l-k-j-i`
    will be determined by their string order.

    DihedralTerm allows wildcard(*) for side atoms.
    For sorting purpose, the wildcard will always be the last.
    During force field matching, the terms with wildcard will have lower priority.
    The terms with wildcard will be used only if exact match cannot be found.

    Parameters
    ----------
    type1 : str
    type2 : str
    type3 : str
    type4 : str

    Attributes
    ----------
    type1 : str
    type2 : str
    type3 : str
    type4 : str
    '''

    def __init__(self, type1, type2, type3, type4):
        super().__init__()
        if '*' in [type2, type3]:
            raise Exception('Wildcard not allowed for center atoms in DihedralTerm')
        if [type1, type4].count('*') in [0, 2]:
            at1, at2, at3, at4 = min([(type1, type2, type3, type4), (type4, type3, type2, type1)])
        elif type1 == '*':
            at1, at2, at3, at4 = (type4, type3, type2, type1)
        else:
            at1, at2, at3, at4 = (type1, type2, type3, type4)
        self.type1 = at1
        self.type2 = at2
        self.type3 = at3
        self.type4 = at4

    @property
    def name(self):
        return '%s,%s,%s,%s' % (self.type1, self.type2, self.type3, self.type4)

    def __lt__(self, other):
        return [self.type1, self.type2, self.type3, self.type4] \
               < [other.type1, other.type2, other.type3, other.type4]

    def __gt__(self, other):
        return [self.type1, self.type2, self.type3, self.type4] \
               > [other.type1, other.type2, other.type3, other.type4]

    @property
    def is_zero(self):
        '''
        Whether or not this dihedral terms always give zero energy.

        For linear groups like alkyne and nitrile, dihedral terms are presented in topology.
        But the parameters for these terms usually equal to zero.

        Returns
        -------
        is :  bool
        '''
        return False


class ImproperTerm(FFTerm):
    '''
    Base class for all improper terms.

    Subclasses should be implemented for each functional form,
    e.g. :class:`PeriodicImproperTerm`.

    During the initialization, the three side atom types will be sorted by their string.

    Improper term is usually used to keep 3-coordinated structures in the same plane.
    The first atom is the central atom, following the convention of GROMACS and CHARMM.
    However, the definition of the value of improper i-j-k-l is vague.
    CHARMM defines the value of the improper as the angle between plane i-j-k and j-k-l.
    Whereas OPLS defines the value as the angle between j-k-i and k-i-l.

    ImproperTerm allows wildcard(*) for side atoms.
    During force field matching, the terms with wildcard will have lower priority.
    The terms with wildcard will be used only if exact match cannot be found.

    Parameters
    ----------
    type1 : str
    type2 : str
    type3 : str
    type4 : str

    Attributes
    ----------
    type1 : str
    type2 : str
    type3 : str
    type4 : str
    '''

    def __init__(self, type1, type2, type3, type4):
        super().__init__()
        if type1 == '*':
            raise Exception('Wildcard not allowed for center atoms in ImproperTerm')
        non_wildcard = [t for t in (type2, type3, type4) if t != '*']
        at2, at3, at4 = list(sorted(non_wildcard)) + ['*'] * (3 - len(non_wildcard))
        self.type1 = type1
        self.type2 = at2
        self.type3 = at3
        self.type4 = at4

    @property
    def name(self):
        return '%s,%s,%s,%s' % (self.type1, self.type2, self.type3, self.type4)

    def __lt__(self, other):
        return [self.type1, self.type2, self.type3, self.type4] \
               < [other.type1, other.type2, other.type3, other.type4]

    def __gt__(self, other):
        return [self.type1, self.type2, self.type3, self.type4] \
               > [other.type1, other.type2, other.type3, other.type4]


class PolarizableTerm(FFTerm):
    '''
    Base class for all polarization terms.

    Subclasses should be implemented for each functional form,
    e.g. :class:`DrudeTerm`.

    Parameters
    ----------
    type : str

    Attributes
    ----------
    type : str
    '''

    def __init__(self, type):
        super().__init__()
        self.type = type

    @property
    def name(self):
        return self.type

    def __lt__(self, other):
        return self.type < other.type

    def __gt__(self, other):
        return self.type > other.type


class LJ126Term(VdwTerm):
    '''
    vdW term in LJ-12-6 function form.

    The energy function is

    >>> U = 4*epsilon*((sigma/r)^12-(sigma/r)^6)

    LJ126 is commonly used, therefore we don't generalize it with :class:`MieTerm`.

    During the initialization, the two atom types will be sorted by their string.

    Parameters
    ----------
    type1 : str
    type2 : str
    epsilon : float
    sigma : float

    Attributes
    ----------
    type1 : str
    type2 : str
    epsilon : float
    sigma : float

    '''

    _zfp_attrs = {
        'type1'  : str,
        'type2'  : str,
        'epsilon': float,
        'sigma'  : float,
    }

    def __init__(self, type1, type2, epsilon, sigma):
        super().__init__(type1, type2)
        self.epsilon = epsilon
        self.sigma = sigma

    def evaluate_energy(self, r):
        return 4 * self.epsilon * ((self.sigma / r) ** 12 - (self.sigma / r) ** 6)


FFTermFactory.register(LJ126Term)


class MieTerm(VdwTerm):
    '''
    vdW term with generalized LJ function form.

    The energy function is

    >>> U = C * epsilon * ((sigma/r)^n - (sigma/r)^m)
    >>> C = n/(n-m) * (n/m)^(m/(n-m))
    >>> r_min = (n/m)^(1/(n-m)) * sigma

    This term is mainly used for Coarse-Grained force field.

    During the initialization, the two atom types will be sorted by their string.

    Parameters
    ----------
    type1 : str
    type2 : str
    epsilon : float
    sigma : float
    repulsion : float
    attraction : float

    Attributes
    ----------
    type1 : str
    type2 : str
    epsilon : float
    sigma : float
    repulsion : float
    attraction : float
    '''

    _zfp_attrs = {
        'type1'     : str,
        'type2'     : str,
        'epsilon'   : float,
        'sigma'     : float,
        'repulsion' : float,
        'attraction': float,
    }

    def __init__(self, type1, type2, epsilon, sigma, repulsion, attraction):
        super().__init__(type1, type2)
        self.epsilon = epsilon
        self.sigma = sigma
        self.repulsion = repulsion
        self.attraction = attraction

    def evaluate_energy(self, r):
        return self.factor_energy() * self.epsilon \
               * ((self.sigma / r) ** self.repulsion - (self.sigma / r) ** self.attraction)

    def factor_energy(self):
        '''
        Get the energy pre-factor of the Mie function.

        >>> C = n/(n-m) * (n/m)^(m/(n-m))

        Returns
        -------
        factor : float
        '''
        rep, att = self.repulsion, self.attraction
        return rep / (rep - att) * (rep / att) ** (att / (rep - att))

    def factor_r_min(self):
        '''
        Get the converting factor from sigma to the distance of energy minimum.

        >>> f = (n/m)^(1/(n-m))

        Returns
        -------
        factor : float
        '''
        rep, att = self.repulsion, self.attraction
        return (rep / att) ** (1 / (rep - att))

    @property
    def is_sdk(self):
        '''
        SDK CGFF use LJ-9-6 and LJ-12-4

        Returns
        -------
        is : bool
        '''
        if self.repulsion == 9 and self.attraction == 6:
            return True
        if self.repulsion == 12 and self.attraction == 4:
            return True
        return False


FFTermFactory.register(MieTerm)


class HarmonicBondTerm(BondTerm):
    '''
    Bond term in harmonic function form

    The energy function is

    >>> U = k * (b-b0)^2

    During the initialization, the two atom types will be sorted by their string.

    Parameters
    ----------
    type1 : str
    type2 : str
    length : float
    k : float
    fixed : bool

    Attributes
    ----------
    type1 : str
    type2 : str
    length : float
    k : float
    fixed : bool
    '''

    _zfp_attrs = {
        'type1' : str,
        'type2' : str,
        'length': float,
        'k'     : float,
        'fixed' : lambda x: bool(strtobool(x))
    }

    def __init__(self, type1, type2, length, k, fixed=False):
        super().__init__(type1, type2, length, fixed=fixed)
        self.k = k

    def evaluate_energy(self, val):
        return self.k * (val - self.length) ** 2


FFTermFactory.register(HarmonicBondTerm)


class MorseBondTerm(BondTerm):
    '''
    Bond term in Morse function form

    The energy function is

    >>> U = depth*(1-exp(-alpha*(r-b0)))^2
    >>> alpha = (k / depth) ** 0.5

    The `k` corresponds to the force constant `k` in HarmonicBondTerm.

    During the initialization, the two atom types will be sorted by their string.

    Parameters
    ----------
    type1 : str
    type2 : str
    length : float
    k : float
    depth : float

    Attributes
    ----------
    type1 : str
    type2 : str
    length : float
    k : float
    depth : float
    '''

    _zfp_attrs = {
        'type1' : str,
        'type2' : str,
        'length': float,
        'k'     : float,
        'depth' : float,
    }

    def __init__(self, type1, type2, length, k, depth):
        super().__init__(type1, type2, length)
        self.k = k
        self.depth = depth

    def evaluate_energy(self, val):
        alpha = (self.k / self.depth) ** 0.5
        return self.depth * (1 - math.exp(-alpha * (val - self.length))) ** 2


FFTermFactory.register(MorseBondTerm)


class HarmonicAngleTerm(AngleTerm):
    '''
    Angle term in harmonic function form

    The energy function is

    >>> U = k * (theta-theta0)^2

    The angle is in unit of radian, and the force constant is in unit of kJ/mol/rad^2.

    During the initialization, the two side atom types will be sorted by their string.

    Parameters
    ----------
    type1 : str
    type2 : str
    type3 : str
    theta : float
    k : float
    fixed : bool

    Attributes
    ----------
    type1 : str
    type2 : str
    type3 : str
    theta : float
    k : float
    fixed : bool
    '''

    _zfp_attrs = {
        'type1': str,
        'type2': str,
        'type3': str,
        'theta': float,
        'k'    : float,
        'fixed': lambda x: bool(strtobool(x))
    }

    def __init__(self, type1, type2, type3, theta, k, fixed=False):
        super().__init__(type1, type2, type3, theta, fixed=fixed)
        self.k = k

    def evaluate_energy(self, val):
        return self.k * (val - self.theta) ** 2

    @property
    def is_linear(self):
        return self.theta > 175 * DEG2RAD


FFTermFactory.register(HarmonicAngleTerm)


class SDKAngleTerm(AngleTerm):
    '''
    Angle term in SDK function form

    The energy function is

    >>> U = k * (theta-theta0)^2 + LJ96
    >>> LJ96 = 6.75 * epsilon * ((sigma/r)^9 - (sigma/r)^6) + epsilon, r < 1.144714 * sigma

    The angle is in unit of radian, and the force constant is in unit of kJ/mol/rad^2.

    This term itself does not contain parameters `epsilon` and `sigma`.
    Instead, it expects a :class:`MieTerm` between `type1` and `type3` existing in the :class:`ForceField` object.
    If such a MieTerm does not exist, or the repulsion and attraction of this MieTerm do not equal to 9 and 6,
    then the ForceField is invalid and cannot be used to construct a :class:`~mstk.simsys.System`.

    During the initialization, the two side atom types will be sorted by their string.

    Parameters
    ----------
    type1 : str
    type2 : str
    type3 : str
    theta : float
    k : float

    Attributes
    ----------
    type1 : str
    type2 : str
    type3 : str
    theta : float
    k : float
    fixed : bool
    '''

    _zfp_attrs = {
        'type1': str,
        'type2': str,
        'type3': str,
        'theta': float,
        'k'    : float,
    }

    def __init__(self, type1, type2, type3, theta, k):
        super().__init__(type1, type2, type3, theta, fixed=False)
        self.k = k

    @property
    def is_linear(self):
        return self.theta > 175 * DEG2RAD


FFTermFactory.register(SDKAngleTerm)


class LinearAngleTerm(AngleTerm):
    '''
    Linear angle described by the distance between the center atom and its equilibrated position

    For angle i-j-k, the energy function is

    >>> U = k_linear * (r_j - r_j_0)^2
    >>> r_j_0 = a * r_i + (1-a) * r_k
    >>> a = b_jk / (b_ij + b_jk)

    The k_linear can be converted from the k_theta in HarmonicAngleTerm

    >>> k_linear = k_theta * 2 * (b_ij + b_jk)^2 / (b_ij * b_jk)^2

    To be consistent with other angle terms, the k_theta is defined as the force constant, in unit of kJ/mol/rad^2

    This term itself does not contain bond length parameters for i-j and j-k bonds
    Instead, it expects two :class:`BondTerm` between each side atom type and `type2` existing in the :class:`ForceField` object.
    If such a `BondTerm` does not exist, then the ForceField is invalid and cannot be used to construct a :class:`~mstk.simsys.System`.

    During the initialization, the two side atom types will be sorted by their string.

    Parameters
    ----------
    type1 : str
    type2 : str
    type3 : str
    k : float

    Attributes
    ----------
    type1 : str
    type2 : str
    type3 : str
    theta : float
    k : float
    fixed : bool
    '''

    _zfp_attrs = {
        'type1': str,
        'type2': str,
        'type3': str,
        'k'    : float,
    }

    def __init__(self, type1, type2, type3, k):
        super().__init__(type1, type2, type3, PI, fixed=False)
        self.k = k

    def calc_a_k_linear(self, b12, b23):
        a = b23 / (b12 + b23)
        k_linear = self.k * 2 * (b12 + b23) ** 2 / (b12 * b23) ** 2
        return a, k_linear

    @property
    def is_linear(self):
        return True


FFTermFactory.register(LinearAngleTerm)


class OplsDihedralTerm(DihedralTerm):
    '''
    Dihedral term in OPLS form

    The energy function is

    >>> U = k_1*(1+cos(phi)) + k_2*(1-cos(2*phi)) + k_3*(1+cos(3*phi)) - k_4*(1-cos(4*phi))

    OPLS form is commonly used, therefore we don't generalize it with :class:`PeriodicDihedralTerm`.

    During the initialization, the sequence of atom types to use `i-j-k-l` or `l-k-j-i`
    will be determined by their string order.

    DihedralTerm allows wildcard(*) for side atoms.
    For sorting purpose, the wildcard will always be the last.
    During force field matching, the terms with wildcard will have lower priority.
    The terms with wildcard will be used only if exact match cannot be found.

    Parameters
    ----------
    type1 : str
    type2 : str
    type3 : str
    type4 : str
    k1 : float
    k2 : float
    k3 : float
    k4 : float

    Attributes
    ----------
    type1 : str
    type2 : str
    type3 : str
    type4 : str
    k1 : float
    k2 : float
    k3 : float
    k4 : float
    '''

    _zfp_attrs = {
        'type1': str,
        'type2': str,
        'type3': str,
        'type4': str,
        'k1'   : float,
        'k2'   : float,
        'k3'   : float,
        'k4'   : float,
    }

    def __init__(self, type1, type2, type3, type4, k1, k2, k3, k4):
        super().__init__(type1, type2, type3, type4)
        self.k1 = k1
        self.k2 = k2
        self.k3 = k3
        self.k4 = k4

    def evaluate_energy(self, val):
        energy = 0
        energy += self.k1 * (1 + math.cos(val)) \
                  + self.k2 * (1 - math.cos(2 * val)) \
                  + self.k3 * (1 + math.cos(3 * val)) \
                  + self.k4 * (1 - math.cos(4 * val))
        return energy

    @property
    def is_zero(self):
        return self.k1 == 0 and self.k2 == 0 and self.k3 == 0 and self.k4 == 0

    def to_periodic_term(self):
        '''
        Get the equivalent PeriodicDihedralTerm

        Returns
        -------
        term : PeriodicDihedralTerm
        '''
        term = PeriodicDihedralTerm(self.type1, self.type2, self.type3, self.type4)
        if self.k1 != 0:
            term.add_phase(0., self.k1, 1)
        if self.k2 != 0:
            term.add_phase(PI, self.k2, 2)
        if self.k3 != 0:
            term.add_phase(0., self.k3, 3)
        if self.k4 != 0:
            term.add_phase(PI, self.k4, 4)

        return term


FFTermFactory.register(OplsDihedralTerm)


class PeriodicDihedralTerm(DihedralTerm):
    '''
    Dihedral term in periodic cosine function form

    The energy function is

    >>> U = \sum^n k_n*(1+cos(n*phi-phi0_n))

    During the initialization, the sequence of atom types to use `i-j-k-l` or `l-k-j-i`
    will be determined by their string order.

    DihedralTerm allows wildcard(*) for side atoms.
    For sorting purpose, the wildcard will always be the last.
    During force field matching, the terms with wildcard will have lower priority.
    The terms with wildcard will be used only if exact match cannot be found.

    PeriodicTermDihedral is initialized without phases.
    The parameters for phases should be added by calling :func:`add_phase`.

    Examples
    --------

    >>> term = PeriodicDihedralTerm('h_1', 'c_4', 'c_4', 'h_1')
    >>> term.add_phase(0.0, 0.6485, 1)
    >>> term.add_phase(3.14, 1.0678, 2)
    >>> term.add_phase(0.0, 0.6226, 3)
    >>> for phase in term.phases:
    >>>     print(phase.phi, phase.k, phase.n)
        0.0 0.6485 1
        3.14 1.0678 2
        0.0 0.6226 3
    >>> term = term.to_opls_term()
    >>> print((term.k1, term.k2, term.k3, term.k4))
        (0.6485 1.0678 0.6226 0.0)

    Parameters
    ----------
    type1 : str
    type2 : str
    type3 : str
    type4 : str

    Attributes
    ----------
    type1 : str
    type2 : str
    type3 : str
    type4 : str
    phases : list of namedtuple
        Each element is a namedtuple :attr:`Phase` describing the parameters for each phase.
    '''

    _zfp_attrs = {
        'type1': str,
        'type2': str,
        'type3': str,
        'type4': str,
    }

    class Phase:
        def __init__(self, phi, k, n):
            if phi < -PI or phi > PI:
                logger.error(f'phi should be in the range of -PI to PI')
            self.phi = phi
            self.k = k
            self.n = n

    def __init__(self, type1, type2, type3, type4):
        super().__init__(type1, type2, type3, type4)
        self.phases = []

    def add_phase(self, phi, k, n):
        '''
        Add parameters for a phase to this dihedral term.

        Multiplicity must be a positive integer. Otherwise, an Exception will be raised.
        If there already exists parameter for this multiplicity, an Exception will be raised.

        Parameters
        ----------
        phi : float
            Phase shift for this multiplicity, in unit of radian
        k : float
            Force constant
        n : int
            Multiplicity
        '''
        if not isinstance(n, int) or n < 1:
            raise Exception('Multiplicity should be a positive integer: %s' % self.name)
        for para in self.phases:
            if para.n == n:
                raise Exception('Duplicated multiplicity: %s' % self.name)
        self.phases.append(PeriodicDihedralTerm.Phase(phi=phi, k=k, n=n))
        self.phases.sort(key=lambda x: x.n)

    def _to_zfp_extra(self) -> {str: str}:
        d = {}
        for phase in self.phases:
            d.update({
                'phi_%i' % phase.n: '%.1f' % (phase.phi * RAD2DEG),  # the angle in ZFP or ZFF file is in unit of degree
                'k_%i' % phase.n  : '%.4f' % phase.k,
            })
        return d

    def _from_zfp_extra(self, d: {str: str}):
        for key in d.keys():
            if key.startswith('phi_'):
                n = int(key.split('_')[-1])
                phi = float(d['phi_%i' % n]) * DEG2RAD  # the angle in ZFP or ZFF file is in unit of degree
                k = float(d['k_%i' % n])
                if k != 0:
                    self.add_phase(phi, k, n)
        if not self.phases:
            self.add_phase(0.0, 0.0, 1)

    def to_zff(self):
        line = '%-16s %-9s %-9s %-9s %-9s' % (self.__class__.get_alias(),
                                              self.type1, self.type2, self.type3, self.type4)
        for phase in self.phases:
            # the angle in ZFP or ZFF file is in unit of degree
            line += ' %9.1f %9.4f %2i' % (phase.phi * RAD2DEG, phase.k, phase.n)
        return line

    def _from_zff_extra(self, line):
        words = line.strip().split()
        str_vals = words[5:]
        if len(str_vals) % 3 != 0:
            raise Exception(f'Invalid parameters for {self.__class__.__name__}')
        for i in range(len(str_vals) // 3):
            phi = float(str_vals[i * 3]) * DEG2RAD  # the angle in ZFP or ZFF file is in unit of degree
            k = float(str_vals[i * 3 + 1])
            n = int(str_vals[i * 3 + 2])
            if k != 0:
                self.add_phase(phi, k, n)
        if not self.phases:
            self.add_phase(0.0, 0.0, 1)

    def evaluate_energy(self, val):
        energy = 0
        for para in self.phases:
            energy += para.k * (1 + math.cos(para.n * val - para.phi))
        return energy

    @property
    def is_zero(self):
        for para in self.phases:
            if para.k != 0:
                return False
        return True

    def to_opls_term(self):
        '''
        Get the equivalent OplsDihedralTerm.

        In OPLS convention, the largest multiplicity equals to 4.
        The phi_0 for multiplicity (1, 2, 3, 4) equal to (0, PI, 0, PI), respectively.
        If it does not follow OPLS convention, an Exception will be raised.

        Returns
        -------
        term : OplsDihedralTerm
        '''
        k1 = k2 = k3 = k4 = 0.0
        for para in self.phases:
            if para.n == 1:
                if para.phi != 0:
                    raise Exception(f'{str(self)} does not follow OPLS convention, phi_1 != 0')
                k1 = para.k
            elif para.n == 2:
                if abs(para.phi - PI) > 0.01:
                    raise Exception(f'{str(self)} does not follow OPLS convention, phi_2 != PI')
                k2 = para.k
            elif para.n == 3:
                if para.phi != 0:
                    raise Exception(f'{str(self)} does not follow OPLS convention, phi_3 != 0')
                k3 = para.k
            elif para.n == 4:
                if abs(para.phi - PI) > 0.01:
                    raise Exception(f'{str(self)} does not follow OPLS convention, phi_4 != PI')
                k4 = para.k
            else:
                raise Exception(f'{str(self)} does not follow OPLS convention, n > 4')

        return OplsDihedralTerm(self.type1, self.type2, self.type3, self.type4, k1, k2, k3, k4)


FFTermFactory.register(PeriodicDihedralTerm)


class OplsImproperTerm(ImproperTerm):
    '''
    Improper term in OPLS cosine function form

    The energy function is

    >>> U = k * (1-cos(2*phi))

    During the initialization, the three side atom types will be sorted by their string.

    ImproperTerm allows wildcard(*) for side atoms.
    During force field matching, the terms with wildcard will have lower priority.
    The terms with wildcard will be used only if exact match cannot be found.

    Care must be taken when exporting OplsImproperTerm.
    For improper a1-a2-a3-a4, a1 is the center atom.
    In OPLS convention, the value of the improper is defined as the angle between plane a2-a3-a1 and a3-a1-a4.

    Parameters
    ----------
    type1 : str
    type2 : str
    type3 : str
    type4 : str
    k : float

    Attributes
    ----------
    type1 : str
    type2 : str
    type3 : str
    type4 : str
    k : float
    '''

    _zfp_attrs = {
        'type1': str,
        'type2': str,
        'type3': str,
        'type4': str,
        'k'    : float,
    }

    def __init__(self, type1, type2, type3, type4, k):
        super().__init__(type1, type2, type3, type4)
        self.k = k

    def evaluate_energy(self, val):
        return self.k * (1 - math.cos(2 * val * DEG2RAD))


FFTermFactory.register(OplsImproperTerm)


class HarmonicImproperTerm(ImproperTerm):
    '''
    Improper term in harmonic function form. Mainly used by CHARMM.

    The energy function is

    >>> U = k * (phi-phi0)^2

    The improper is in unit of radian, and the force constant is in unit of kJ/mol/rad^2.

    During the initialization, the three side atom types will be sorted by their string.

    ImproperTerm allows wildcard(*) for side atoms.
    During force field matching, the terms with wildcard will have lower priority.
    The terms with wildcard will be used only if exact match cannot be found.

    Care must be taken when exporting HarmonicImproperTerm.
    For improper a1-a2-a3-a4, a1 is the center atom.
    In CHARMM convention, the improper is defined as the angle between plane a1-a2-a3 and a2-a3-a4.

    Parameters
    ----------
    type1 : str
    type2 : str
    type3 : str
    type4 : str
    phi : float
    k : float

    Attributes
    ----------
    type1 : str
    type2 : str
    type3 : str
    type4 : str
    phi : float
    k : float
    '''

    _zfp_attrs = {
        'type1': str,
        'type2': str,
        'type3': str,
        'type4': str,
        'phi'  : float,
        'k'    : float,
    }

    # the angle in ZFP or ZFF file is in unit of degree. one digit is enough
    _zfp_convert = {
        'phi': (lambda x: x * DEG2RAD, lambda x: round(x * RAD2DEG, 2))
    }

    def __init__(self, type1, type2, type3, type4, phi, k):
        super().__init__(type1, type2, type3, type4)
        self.phi = phi
        self.k = k

    def evaluate_energy(self, val):
        return self.k * ((val - self.phi) * DEG2RAD) ** 2


FFTermFactory.register(HarmonicImproperTerm)


class DrudeTerm(PolarizableTerm):
    '''
    Polarization term described by Drude induced dipole.

    Currently, only the isotropic Drude polarization is implemented.
    The energy function for polarization is

    >>> E = k * d^2

    The charge assigned to the Drude particle is

    >>> q = - sqrt(4 * pi * eps_0 * (2k) * alpha)

    If this atom is bonded with hydrogen atoms, the argument `merge_alpha_H` determines the polarizability of neighbour H atoms.
    The alpha of these H atoms will be merged into the this atom.
    Therefore, DrudeTerm for H atom types should not be provided explicitly. Otherwise, there will be double count.
    The H atoms are identified by check the :attr:`~mstk.topology.Atom.symbol` attribute.
    An atom is considered to be hydrogen if symbol equals 'H'.

    Parameters
    ----------
    type : str
    alpha : float
    thole : float
    k : float
    mass : float
    merge_alpha_H : float

    Attributes
    ----------
    type : str
    alpha : float
    thole : float
    k : float
    mass : float
    merge_alpha_H : float
    '''

    _zfp_attrs = {
        'type'         : str,
        'alpha'        : float,
        'thole'        : float,
        'k'            : float,
        'mass'         : float,
        'merge_alpha_H': float,
    }

    def __init__(self, type: str, alpha, thole, k=4184 / 2 * 100, mass=0.4, merge_alpha_H=0.0):
        super().__init__(type)
        self.alpha = alpha  # nm^3
        self.thole = thole
        self.k = k  # kJ/mol/nm^2
        self.mass = mass
        self.merge_alpha_H = merge_alpha_H

    def get_charge(self, alpha=None):
        '''
        Get the charge offset between parent atom and Drude particle.

        The charge assigned on the Drude particle is the negative of this charge offset.

        Parameters
        ----------
        alpha : float
            The polarizability of parent atom in unit of nm^3.

        Returns
        -------
        charge : float
            The charge offset from Drude particle to parent atom.
        '''
        if alpha is None:
            alpha = self.alpha
        factor = math.sqrt(1E-6 / AVOGADRO) / ELEMENTARY_CHARGE
        return math.sqrt(4 * PI * VACUUM_PERMITTIVITY * (2 * self.k) * alpha) * factor


FFTermFactory.register(DrudeTerm)


class VirtualSiteTerm(FFTerm):
    '''
    Base class for all virtual site terms.

    Subclasses should be implemented for each functional form,
    e.g. :class:`TIP4PTerm`.

    Parameters
    ----------
    type : str

    Attributes
    ----------
    type : str
    '''

    def __init__(self, type):
        super().__init__()
        self.type = type

    @property
    def name(self):
        return self.type

    def __lt__(self, other):
        return self.type < other.type

    def __gt__(self, other):
        return self.type > other.type


class TIP4PSiteTerm(VirtualSiteTerm):
    '''
    TIP4P style virtual site term.

    Parameters
    ----------
    type : str
    type_O : str
    type_H : str
    d : float

    Attributes
    ----------
    type : str
    type_O : str
    type_H : str
    d : float
    '''

    _zfp_attrs = {
        'type'  : str,
        'type_O': str,
        'type_H': str,
        'd'     : float,
    }

    def __init__(self, type: str, type_O: str, type_H: str, d: float):
        super().__init__(type)
        self.type_O = type_O
        self.type_H = type_H
        self.d = d


FFTermFactory.register(TIP4PSiteTerm)



import numpy as np
import openmm.openmm as mm
from openmm.app import Simulation
from ..unit import *

class DrudeTemperatureReporter(object):
    '''
    DrudeTemperatureReporter reports the temperatures of different DOFs in a Drude simulation system.

    The temperatures for three sets of degrees of freedom are reported
    -- molecular center of mass, internal atomic and Drude temperature.
    It's better to set the reportInterval larger than 10000 to avoid performance penalty

    Parameters
    ----------
    file : string
        The file to write to
    reportInterval : int
        The interval (in time steps) at which to write frames
    append : bool
        Whether or not append to existing file
    '''

    def __init__(self, file, reportInterval, append=False):
        self._reportInterval = reportInterval
        if append:
            self._out = open(file, 'a')
        else:
            self._out = open(file, 'w')
        self._hasInitialized = False

    def describeNextReport(self, simulation):
        """Get information about the next report this object will generate.

        Parameters
        ----------
        simulation : Simulation
            The Simulation to generate a report for

        Returns
        -------
        tuple
            A six element tuple. The first element is the number of steps
            until the next report. The next four elements specify whether
            that report will require positions, velocities, forces, and
            energies respectively.  The final element specifies whether
            positions should be wrapped to lie in a single periodic box.
        """
        steps = self._reportInterval - simulation.currentStep%self._reportInterval
        return (steps, False, True, False, False)

    def report(self, simulation, state):
        """Generate a report.

        Parameters
        ----------
        simulation : Simulation
            The Simulation to generate a report for
        state : State
            The current state of the simulation
        """
        system: mm.System = simulation.system
        if not self._hasInitialized:
            self.n_atom = system.getNumParticles()
            self.molecules: [[int]] = [list(atoms) for atoms in simulation.context.getMolecules()]
            self.n_mol = len(self.molecules)
            self.mol_atoms = np.zeros(self.n_atom, dtype=int)  # record which molecule the atoms are in
            self.mass_molecules = np.zeros(self.n_mol) # record the mass of molecules
            for i, atoms in enumerate(self.molecules):
                for atom in atoms:
                    self.mol_atoms[atom] = i
                    self.mass_molecules[i] += system.getParticleMass(atom).value_in_unit(unit.dalton)

            self.dof_com = np.count_nonzero(self.mass_molecules) * 3
            self.dof_atom = self.dof_drude = 0
            for i in range(self.n_atom):
                if system.getParticleMass(i) > 0 * unit.dalton:
                    self.dof_atom += 3
            self.dof_atom -= (self.dof_com + system.getNumConstraints())

            if any(type(f) == mm.CMMotionRemover for f in system.getForces()):
                self.dof_com -= 3

            drude_set = set()
            self.pair_set = set()
            force = next(f for f in system.getForces() if type(f) == mm.DrudeForce)
            self.dof_atom -= 3 * force.getNumParticles()
            self.dof_drude += 3 * force.getNumParticles()
            for i in range(force.getNumParticles()):
                i_drude, i_core = force.getParticleParameters(i)[:2]
                drude_set.add(i_drude)
                self.pair_set.add((i_drude, i_core))
            self.drude_array = np.array(list(drude_set))
            self.atom_array = np.array(list(set(range(self.n_atom)) - drude_set))

            self._hasInitialized = True
            print('#"Step"\t"T_COM"\t"T_Atom"\t"T_Drude"\t"KE_COM"\t"KE_Atom"\t"KE_Drude"', file=self._out)

        velocities = state.getVelocities(asNumpy=True).value_in_unit(unit.nanometer / unit.picosecond)
        masses = np.array([system.getParticleMass(i).value_in_unit(unit.dalton) for i in range(self.n_atom)])

        vel_mol = np.zeros([self.n_mol, 3])
        for i, atoms in enumerate(self.molecules):
            if self.mass_molecules[i] == 0:
                continue
            mv = masses[atoms][:, np.newaxis] * velocities[atoms]
            vel_mol[i] = np.sum(mv, axis=0) / self.mass_molecules[i]
        mvv_com = self.mass_molecules * np.sum(vel_mol ** 2, axis=1)
        ke_com = mvv_com.sum() / 2 * (unit.nanometer / unit.picosecond) ** 2 * unit.dalton
        t_com = (2 * ke_com / (self.dof_com * unit.MOLAR_GAS_CONSTANT_R))

        velocities -= np.array([vel_mol[self.mol_atoms[i]] for i in range(self.n_atom)])
        for i_drude, i_core in self.pair_set:
            v_drude = velocities[i_drude]
            v_core = velocities[i_core]
            m_drude = masses[i_drude]
            m_core = masses[i_core]
            m_com = m_drude + m_core
            m_rel = m_drude * m_core / m_com
            v_com = (m_drude * v_drude + m_core * v_core) / m_com
            v_rel = v_drude - v_core
            velocities[i_drude] = v_rel
            velocities[i_core] = v_com
            masses[i_drude] = m_rel
            masses[i_core] = m_com
        mvv = masses * np.sum(velocities ** 2, axis=1)
        ke = mvv[self.atom_array].sum() / 2 * (unit.nanometer / unit.picosecond) ** 2 * unit.dalton
        ke_drude = mvv[self.drude_array].sum() / 2 * (unit.nanometer / unit.picosecond) ** 2 * unit.dalton
        t = (2 * ke / (self.dof_atom * unit.MOLAR_GAS_CONSTANT_R))
        t_drude = (2 * ke_drude / (self.dof_drude * unit.MOLAR_GAS_CONSTANT_R))
        print(simulation.currentStep,
              t_com.value_in_unit(kelvin), t.value_in_unit(kelvin), t_drude.value_in_unit(kelvin),
              ke_com.value_in_unit(kJ_mol), ke.value_in_unit(kJ_mol), ke_drude.value_in_unit(kJ_mol),
              sep='\t', file=self._out)

        if hasattr(self._out, 'flush') and callable(self._out.flush):
            self._out.flush()

    def __del__(self):
        self._out.close()



_atomic_number = {'H': 1, 'He': 2, 'Li': 3, 'Be': 4, 'B': 5,
                  'C': 6, 'N': 7, 'O': 8, 'F': 9, 'Ne': 10,
                  'Na': 11, 'Mg': 12, 'Al': 13, 'Si': 14, 'P': 15,
                  'S': 16, 'Cl': 17, 'Ar': 18, 'K': 19, 'Ca': 20,
                  'Ti': 22, 'Fe': 26, 'Zn': 30, 'Se': 34, 'Br': 35,
                  'Kr': 36, 'Mo': 42, 'Ru': 44, 'Sn': 50, 'Te': 52,
                  'I': 53, 'Xe': 54, 'UNK': -1, 'DP': -2, 'VS': -3}

_atomic_mass = {'H': 1.008, 'He': 4.003, 'Li': 6.941, 'Be': 9.012, 'B': 10.811,
                'C': 12.011, 'N': 14.006, 'O': 15.999, 'F': 18.998, 'Ne': 20.180,
                'Na': 22.990, 'Mg': 24.305, 'Al': 26.982, 'Si': 28.086, 'P': 30.974,
                'S': 32.065, 'Cl': 35.453, 'Ar': 39.948, 'K': 39.098, 'Ca': 40.078,
                'Ti': 47.867, 'Fe': 55.845, 'Zn': 65.38, 'Se': 78.971, 'Br': 79.904,
                'Kr': 83.798, 'Mo': 95.96, 'Ru': 101.07, 'Sn': 118.710, 'Te': 127.60,
                'I': 126.904, 'Xe': 131.293, 'UNK': 0.0, 'DP': 0.0, 'VS': 0.0}

_atomic_symbol = {v: k for k, v in _atomic_number.items()}


class Element():
    def __init__(self, arg):
        if isinstance(arg, int):
            self.number = arg
            self.symbol = _atomic_symbol[arg]
        elif isinstance(arg, str):
            self.symbol = arg
            self.number = _atomic_number[arg]
        else:
            raise Exception('Element should be initiated with atomic number or symbol')

        self.mass = _atomic_mass[self.symbol]

    def __repr__(self):
        return f'<Element: {self.symbol}>'

    @staticmethod
    def guess_from_atom_type(type):
        '''
        Guess the element from the first two characters of atom type
        Will only guess TEAM and CL&P atom types
        TEAM atom types are lowercase
        CL&P atom types are title case
        '''
        if type[0].islower():
            type = str.upper(type[0]) + type[1:]
        if type[:2] in _atomic_number.keys():
            symbol = type[:2]
        elif type[0] in _atomic_number.keys():
            symbol = type[0]
        else:
            symbol = 'UNK'
        return Element(symbol)



from mstk.trajectory.handler import TrjHandler


class CombinedTrj(TrjHandler):
    '''
    Read several trajectory files at the same time.

    It is useful for processing truncated trajectories after restarting simulation.
    All trajectory files should contain the same number of atoms.
    '''

    def __init__(self, files, mode='r'):
        if mode != 'r':
            raise Exception('CombinedTrj is read-only')

        super().__init__()

        self._handlers = []
        for file in files:
            Handler = TrjHandler.get_handler_for_file(file)
            self._handlers.append(Handler(file, mode='r'))

    def get_info(self):
        self._i_frame_handler = []
        self._i_frame_offset = []
        for handler in self._handlers:
            handler.get_info()
            self._i_frame_handler += [handler] * handler.n_frame
            self._i_frame_offset += list(range(handler.n_frame))

        if set([handler.n_atom for handler in self._handlers]).__len__() != 1:
            raise Exception('All trajectories should have same number of atoms')

        self.n_frame = len(self._i_frame_offset)
        self.n_atom = self._handlers[0].n_atom
        return self.n_atom, self.n_frame

    def close(self):
        for handler in self._handlers:
            handler.close()

    def read_frame(self, i_frame, frame):
        handler = self._i_frame_handler[i_frame]
        i = self._i_frame_offset[i_frame]
        handler.read_frame(i, frame)



""" One dimension canny algorithm, for detecting edges
"""
# Last modified: 2017.2

import math
import numpy as np


def canny1d(interval, y, nms_width, peak_threshold=0.1, gauss_sigma=1, gauss_width=5, grad_width=3, debug=False):
    """ 1-dimension canny algorithm. 
     Arguments:
        interval: time interval;
        y: numpy array for raw data;
        gauss_sigma: Gaussian filter sigma, default is 0.35, determined by test, no modify recommended;
        guass_width: Gaussian filter width, default is 5;
        nms_width: Non-maximum supression width. It is the scan width of window in order to delete maximum.
            Modify this only when two much near peaks are found;
        peak_threshold: Peaks with absolute value lower than it will be deleted.
            Not recommend to set larger than 0.3;
        grad_width: Gradient operator width. Higher will make the curve smoother. Less than 5 is recommended.
    Returns:
        A list of tuple (t, value) for peaks;
    """
    if debug:
        try:
            import matplotlib.pyplot as plt
        except:
            print('matplotlib not found, cannot debug')
            debug = False

    y_oper = np.concatenate((y, y))  # stack twice for boundary results
    t = np.linspace(0, interval * len(y), len(y))
    t_oper = np.concatenate((t, t + len(y) * interval))

    if debug:
        plt.figure(1)
        plt.plot(t_oper, y_oper)
    smooth_y = uniform(np.convolve(y_oper, get_gaussian_1d(gauss_sigma, gauss_width), mode='same'))
    if debug:
        plt.plot(t_oper, smooth_y)
    grad_y = np.convolve(smooth_y, get_grad_operator(grad_width), mode='same')
    if debug:
        plt.figure(2)
        plt.plot(t_oper, grad_y)
    grad_y_nms = non_max_suppress(np.abs(grad_y), nms_width)
    if debug:
        plt.figure(2)
        plt.plot(t_oper, grad_y_nms)

    max_peak = np.max(grad_y_nms[len(y) // 2:len(y) * 3 // 2])
    if debug:
        print('max peak', max_peak)

    raw_peak_idx = np.where(grad_y_nms > peak_threshold * max_peak)[0]
    raw_peak_idx = raw_peak_idx[(raw_peak_idx >= len(y) / 2) * (raw_peak_idx < len(y) * 1.5)]

    hpws = get_halfwidth(np.abs(grad_y), raw_peak_idx.tolist())

    peaks = []

    t_oper = np.concatenate((t, t))

    for idx, hpw in zip(raw_peak_idx, hpws):
        peaks.append((t_oper[idx], grad_y[idx], hpw * interval))

    # used in debug
    if debug:
        plt.show()
        print('raw peaks are (total', len(peaks), '):')
        print(peaks)

    return sorted(peaks, key=lambda x: x[0])


def get_gaussian_1d(sigma, width):
    """ return 1d gauss blur kernel. width must be odd."""
    half_width = (width - 1) // 2
    kernel = np.zeros(width)
    for i in range(width):
        kernel[i] = (i - half_width) ** 2
    return np.exp(-kernel / (2 * sigma * sigma)) / (2 * math.pi * sigma * sigma)


def get_grad_operator(width):
    """ return 1d gradient operator. width must be odd
    """
    half_width = (width - 1) // 2
    grad_operator = np.zeros(width)
    for i in range(0, half_width):
        grad_operator[i] = - i - 1

    for i in range(half_width + 1, width):
        grad_operator[i] = width - i
    return grad_operator / (width - 1)


def non_max_suppress(data, width):
    """ set non-local-maximum to 0
    """
    half_width = (width + 1) // 2
    localmax = np.zeros_like(data)
    for i in range(len(data)):
        if data[i] == np.max(data[max(i - half_width, 0):min(i + half_width, len(data))]):
            localmax[i] = data[i]
    return localmax


def get_halfwidth(data, peaks):
    hpw = []
    peaks_tmp = [max(0, peaks[-1] - len(data) // 2)] + peaks + [min(peaks[0] + len(data) // 2, len(data))]
    # print(peaks_tmp)
    for i in range(1, len(peaks_tmp) - 1):
        segment = data[(peaks_tmp[i - 1] + peaks_tmp[i]) // 2:(peaks_tmp[i + 1] + peaks_tmp[i]) // 2]
        # print(peaks_tmp[i], data[peaks_tmp[i]]/2)
        hpw.append(len(segment[segment > data[peaks_tmp[i]] / 2]))
        # print(hpw[-1])
    return hpw


def uniform(array):
    """ Set max to 1, min to 0.
    """
    return (array - np.min(array)) / (np.max(array) - np.min(array))



from openmm import unit as _unit

kelvin = _unit.kelvin
bar = _unit.bar
ps = _unit.picosecond
nm = _unit.nanometer
kJ_mol = _unit.kilojoule_per_mole
kcal_mol = _unit.kilocalorie_per_mole
qe = _unit.elementary_charge
dalton = _unit.dalton
item = _unit.item
meter = _unit.meter
angstrom = _unit.angstrom
volt = _unit.volt
farad = _unit.farad



from mstk.topology import Topology
from mstk.trajectory import Frame
from mstk.trajectory.handler import TrjHandler


class Xyz(TrjHandler):
    '''
    Read and write positions from XYZ file.
    '''

    def __init__(self, file, mode='r'):
        super().__init__()
        if mode not in ('r', 'w', 'a'):
            raise Exception('Invalid mode')

        if mode == 'r':
            # open it in binary mode so that we can correctly seek despite of line ending
            self._file = open(file, 'rb')
        elif mode == 'a':
            self._file = open(file, 'ab')
        elif mode == 'w':
            self._file = open(file, 'wb')

    def get_info(self):
        try:
            self.n_atom = int(self._file.readline())
        except:
            raise Exception('Invalid XYZ file')
        self._file.seek(0)

        # read in the file once and build a list of line offsets
        # the last element is the length of whole file
        self._line_offset = [0]
        offset = 0
        for line in self._file:
            offset += len(line)
            self._line_offset.append(offset)
        self._file.seek(0)

        # build a list of frame offsets
        self.n_frame = len(self._line_offset) // (2 + self.n_atom)
        self._frame_offset = []
        for i in range(self.n_frame + 1):
            line_start = (2 + self.n_atom) * i
            self._frame_offset.append(self._line_offset[line_start])

        return self.n_atom, self.n_frame

    def read_frame(self, i_frame, frame):
        # skip to frame i and read only this frame
        self._file.seek(self._frame_offset[i_frame])
        lines = self._file.read(self._frame_offset[i_frame + 1] - self._frame_offset[i_frame]) \
            .decode().splitlines()
        for i in range(self.n_atom):
            words = lines[i + 2].split()
            x = float(words[1]) / 10  # convert A to nm
            y = float(words[2]) / 10
            z = float(words[3]) / 10
            frame.positions[i][:] = x, y, z

    def write_frame(self, frame, topology, subset=None, **kwargs):
        '''
        Write a frame into the opened XYZ file

        Parameters
        ----------
        frame : Frame
        topology : Topology
        subset : list of int
        kwargs : dict
            Ignored
        '''
        if subset is None:
            subset = list(range(len(frame.positions)))

        string = '%i\n' % len(subset)
        string += 'Created by mstk: step= %i, t= %f ps\n' % (frame.step, frame.time)

        for ii, id in enumerate(subset):
            atom = topology.atoms[id]
            pos = frame.positions[id] * 10  # convert from nm to A
            # atom.symbol is more friendly than atom.type for visualizing trajectory
            string += '%-8s %10.5f %10.5f %10.5f\n' % (atom.symbol, pos[0], pos[1], pos[2])

        self._file.write(string.encode())
        self._file.flush()

TrjHandler.register_format('.xyz', Xyz)



import sys
from openmm.app.gromacsgrofile import GromacsGroFile
from openmm.unit import nanometer, picosecond, norm, is_quantity


class GroFile(GromacsGroFile):
    '''
    GroFile is a parser and writer for Gromacs .gro file.

    GroFile extends GromacsGroFile from OpenMM python API by adding writing ability.

    Parameters
    ----------
    file : string
        the name of the file to load
    '''

    @staticmethod
    def writeFile(topology, positions, vectors, file, time=None, subset=None, velocities=None):
        '''
        Write positions (and optionally velocities) of atoms into a GRO file

        It is possible to write a subset of atoms into the GRO file by providing `subset` argument.

        Parameters
        ----------
        topology : openmm.app.Topology
        positions : array_like of shape (n_atom, 3)
            The length of positions should equal to the number of atoms in the topology, even when subset is provided.
        vectors :  array_like of shape (3, 3)
            The full box vectors.
        file : str or FileIO
        time : float
        subset : list of int, optional
            If not provided, then all atoms will be written.
        velocities array_like of shape (n_atom, 3), optional
            The length of velocities should equal to the number of atoms in the topology, even when subset is provided.
            If not provided, then velocities information will not be written.
        '''
        if type(file) is str:
            _file = open(file, 'w')
        else:
            _file = file

        GroFile._writeHeader(time, _file)
        GroFile._writeModel(topology, positions, _file, subset, velocities)
        GroFile._writeFooter(vectors, _file)

        if type(file) is str:
            _file.close()

    @staticmethod
    def _writeHeader(time=None, file=sys.stdout):
        """Write out the header for a PDB file.

        Parameters
        ----------
        topology : Topology
            The Topology defining the molecular system being written
        file : file=stdout
            A file to write the file to
        """
        if time is None:
            time = 0.0
        elif is_quantity(time):
            time = time.value_in_unit(picosecond)

        print("written by openmm t = %.3f ps" % time, file=file)

    @staticmethod
    def _writeModel(topology, positions, file=sys.stdout, subset=None, velocities=None):
        """Write out a model to a PDB file.

        Parameters
        ----------
        topology : Topology
            The Topology defining the model to write
        positions : list
            The list of atomic positions to write
        file : file=stdout
            A file to write the model to
        subset : list(int)=None
            If not None, only the selected atoms will be written
        """

        atoms = list(topology.atoms())
        if len(atoms) != len(positions):
            raise ValueError('The number of positions must match the number of atoms')
        if is_quantity(positions):
            positions = positions.value_in_unit(nanometer)
        if velocities is not None:
            if len(atoms) != len(velocities):
                raise ValueError('The number of velocities must match the number of atoms')
            if is_quantity(velocities):
                velocities = velocities.value_in_unit(nanometer / picosecond)

        if subset is None:
            subset = list(range(len(atoms)))

        print('%i' % len(subset), file=file)
        for ii, i in enumerate(subset):
            atom = atoms[i]
            residue = atom.residue
            coords = positions[i]
            # writing atom symbol instead of name makes visualization easier
            if atom.element is not None:
                name = atom.element.symbol
            else:
                name = ''.join(i for i in atom.name if not i.isdigit())
            line = '%5i%5s%5s%5i%8.3f%8.3f%8.3f' % (
                (residue.index + 1) % 100000, residue.name[:5], name[:5],
                (atom.index + 1) % 100000, coords[0], coords[1], coords[2])
            if velocities is not None:
                vel = velocities[i]
                line += '%8.4f%8.4f%8.4f' % (vel[0], vel[1], vel[2])
            print(line, file=file)

    @staticmethod
    def _writeFooter(periodicBoxVectors, file=sys.stdout):
        """Write out the footer for a PDB file.

        Parameters
        ----------
        topology : Topology
            The Topology defining the molecular system being written
        file : file=stdout
            A file to write the file to
        """
        vectors = periodicBoxVectors.value_in_unit(nanometer)
        xx, xy, xz = vectors[0]
        yx, yy, yz = vectors[1]
        zx, zy, zz = vectors[2]
        print(' %.3f %.3f %.3f %.3f %.3f %.3f %.3f %.3f %.3f' % (
            xx, yy, zz, xy, xz, yx, yz, zx, zy), file=file)
